# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from datetime import datetime

import pytz

from clusterondemand.images.find import CODImage, ImageSource
from clusterondemandconfig import ConfigNamespace
from cmbrightimagerepo.repository import Repository

imagerepo_ns = ConfigNamespace("imagerepo", help_section="image repository parameters")
imagerepo_ns.add_parameter(
    "root_manifest",
    advanced=True,
    default="http://support.brightcomputing.com/imagerepo/repo.yaml",
    help_varname="URL",
    help="URL to the image repository yaml manifest"
)


class RepoImageSource(ImageSource):
    def __init__(self, root_manifest=None, **kwargs):
        super().__init__(**kwargs)
        self.root_manifest = root_manifest
        assert not self.uuids

    @classmethod
    def from_config(cls, config, ids=None):
        assert ids is None
        return RepoImageSource(
            ids=config["ids"],
            root_manifest=config["root_manifest"],
            tags=config["tags"],
            version=config["version"],
            distro=config["distro"],
            package_groups=config["package_groups"],
            revision=config["revision"],
            status=config["status"],
            advanced=config["advanced"],
            image_visibility=config["image_visibility"],
            cloud_type=config["cloud_type"],
        )

    def _iter_from_source(self):
        repo = Repository.make_from_url(self.root_manifest)
        images = repo.images_rec_noraise()
        return (
            make_cod_image_from_repo(image)
            for image in images
        )


def repo_image_name(image):
    image_type = image.image_type()
    if image_type == "headnode":
        prefix = "bcmh"
    elif image_type == "node":
        prefix = "bcmn"
    elif image_type == "node-installer":
        prefix = "bcni"
    elif image_type == "edge-iso":
        prefix = "bcm-ni-edge"
    else:
        raise Exception("Unsupported image type %s" % (image_type))
    return "{0}-{1}-{2}".format(
        prefix,
        image.id(),
        image.revision()
    )


def make_cod_image_from_repo(image):
    # Remove everything after first digit
    # Same logic as in build-cod-image.sh
    distro_family = re.sub(r"\d.*", "", image.os())

    cod_image = CODImage(
        bcm_optional_info=None,
        cloud_type=image.cloud_type(),
        cmd_revision=image.cmd_revision(),
        created_at=datetime.fromtimestamp(image.created_at(), pytz.utc),
        distro=image.os(),
        distro_family=distro_family,
        id=image.id(),
        name=repo_image_name(image),
        image_visibility="public",
        revision=image.revision(),
        size=image.size(),
        tags=image.tags(),
        type=image.image_type(),
        uploaded_at=None,
        uuid="N/A",
        version=str(image.bcm_version()),
        package_groups=sorted(image.package_groups()),
    )
    cod_image.repo_image = image
    return cod_image



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division

import concurrent.futures
import getpass
import itertools
import logging
import os
import re
import signal
import socket
import string
import sys
import time
import traceback
import yaml
from datetime import datetime, timedelta
from subprocess import CalledProcessError, check_call
from typing import Dict, List

import netaddr
import tenacity
from dateutil.parser import parse
from dateutil.tz import tzlocal, tzutc
from six.moves import input

from clusterondemand.exceptions import CODException
from clusterondemandconfig import ConfigNamespace, config

from .decorators import static_vars

log = logging.getLogger("cluster-on-demand")

DATE_TIME_FORMAT = "%m-%d %H:%M"
MAX_CLUSTER_NAME_LENGTH = 64

# reserved for SIGINT signal handler
MAX_DELAY_SECS = 2
_interrupts = 0
_last_seen = 0

UUID_RE = re.compile(r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$", re.IGNORECASE)

UBUNTU_DEV_REPO = """
deb http://dev.brightcomputing.com/bright/apt/${repo_type}/$$(ARCH)/${bcm_version}/ubuntu/${ubuntu_version}/base/ ./
""".strip()

confirm_ns = ConfigNamespace("confirm", "Confirm operations")
confirm_ns.add_switch_parameter(
    "yes",
    flags=["-y"],
    help="Do not ask for confirmation when executing operations"
)


def is_uuid(string):
    return UUID_RE.match(string) is not None


def valid_cluster_name_regex():
    # Following RFC952 and RFC1123, a cluster name may only consist of alphabetic characters, digits
    # and dashes. It may not start or end with a dash. Additionally, we also don't allow cluster
    # names that consist of only digits.
    # (Note that all these restrictions are also enforced by cmdaemon.)
    only_digits_name = r"^[0-9]+$"
    single_char_name = r"[a-zA-Z]"
    multi_char_name = r"[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9]"

    return r"(?!%s)^(%s|%s)$" % (only_digits_name, single_char_name, multi_char_name)


def is_valid_cluster_name(cluster_name):
    return (
        cluster_name and
        len(cluster_name) <= MAX_CLUSTER_NAME_LENGTH and
        bool(re.match(valid_cluster_name_regex(), cluster_name))
    )


def is_valid_ip_address(ip):
    try:
        netaddr.IPAddress(ip)
        return True
    except (netaddr.core.AddrFormatError, ValueError):
        return False


def is_valid_positive_integer(num):
    try:
        if int(num) <= 0:
            return False
        return True
    except ValueError:
        return False


def is_writable_directory(fullpath):
    return os.path.exists(fullpath) and os.path.isdir(fullpath) and os.access(fullpath, os.W_OK)


def format_timestamp(timestamp, format=DATE_TIME_FORMAT):
    """
    Format timestamp.

    :param date:   UNIX timestamp
    :param format:
    :return:       formatted local date/time
    """
    return datetime.fromtimestamp(timestamp).strftime(format)


def format_to_local_date_time(date, tz=None, format=DATE_TIME_FORMAT):
    """
    Format data to local date/time.

    :param date:   ISO 8601 date  (with or without timezone; no timezone means local time)
    :param tz: tzinfo (default: tzlocal())
    :param format:
    :return:       formatted local date/time
    """
    d = parse(date)
    tz = tz or tzlocal()
    if d.tzinfo is not None:
        d = d.astimezone(tz)
    return d.strftime(format)


def get_time_ago_from_iso(date):
    """
    Return human readable version of how long 'date' was ago.

    :param date: ISO 8601 date  (with or without timezone; no timezone means local time)
    :return:  "1d 12h"    (human string time from now
    """
    d1 = parse(date)
    d2 = datetime.now(d1.tzinfo)

    return get_time_ago(d1, d2)


def get_time_ago(past, now=None):
    """Return human readable version of how long 'past' was ago."""
    N_SEC_PER_DAY = 60 * 60 * 24
    N_SEC_PER_HOUR = 60 * 60
    N_SEC_PER_MINUTE = 60

    now = now or datetime.now(past.tzinfo)
    delta = now - past

    seconds = delta.days * N_SEC_PER_DAY + delta.seconds
    days, seconds = divmod(seconds, N_SEC_PER_DAY)
    hours, seconds = divmod(seconds, N_SEC_PER_HOUR)
    minutes, seconds = divmod(seconds, N_SEC_PER_MINUTE)

    if hours == 23 and 30 < minutes:
        return "%dd 0h" % (days + 1)
    elif days:
        return "%dd %dh" % (days, hours)
    elif 30 < seconds:
        return "%dh %dm" % (hours, minutes + 1)
    else:
        return "%dh %dm" % (hours, minutes)


def setup_signal_handlers():
    def sigint_signal_handler(*args):
        global _interrupts
        global _last_seen

        log.debug("Caught SIGINT!")

        now = time.time()
        if now > _last_seen + MAX_DELAY_SECS:
            _interrupts = 0

        _last_seen = now
        _interrupts += 1
        if _interrupts == 1:
            log.info("\n\n\033[91mPress ctrl+c again, quickly, to abort\033[0m\n")

        elif _interrupts > 1:
            log.info("Aborted by user (ctrl+c).")
            raise KeyboardInterrupt

    signal.signal(signal.SIGINT, sigint_signal_handler)


def confirm(message=""):
    if message:
        log.info(message)

    if config["yes"]:
        log.debug("The option --yes is set. Skipping confirmation...")
        return True

    while True:
        log.info("Proceed? [yes/no]")

        reply = input().lower()
        if reply == "no" or reply == "n":
            log.info("Aborted.")
            return False
        elif reply == "yes" or reply == "y":
            return True
        else:
            log.info("Please type 'yes' or 'no'.")


def confirm_with_enter(message=""):
    if message:
        log.info(message)

    if config["yes"]:
        log.debug("The option --yes is set. Skipping confirmation...")
        return

    while True:
        try:
            log.info("Press ENTER to continue, ctrl+c to abort.")

            reply = input().lower()
            if reply != "" and reply != "q":
                log.info("ENTER to continue, ctrl+c (or type 'q') to abort.")
                continue
            elif reply == "q":
                log.info("Aborted.")
                exit(1)
            else:
                return
        except KeyboardInterrupt:
            log.info("Aborted.")
            exit(1)


def confirm_cluster_creation(num_clusters=1):
    if config["yes"]:
        log.debug("The option --yes is set. Skipping confirmation...")
        return

    while True:
        try:

            if num_clusters == 1:
                log.info("Press ENTER to continue and create the cluster.")
            else:
                log.info("Press ENTER to continue and create %s clusters." % num_clusters)

            log.info("Press ctrl+c (or type 'a') to abort. Type 'i' for more info.")

            reply = input().lower()
            if reply != "" and reply != "a" and reply != "i":
                continue
            elif reply == "a":
                log.info("Aborted.")
                exit(1)
            elif reply == "i":
                log.info("")
                log.info('You can set "ask_to_confirm_cluster_creation: false" in the config file'
                         " to permanently disable this question.")
                log.info("Add '-y' to the command line to skip this question in the future.")
                log.info("")

            else:
                return
        except KeyboardInterrupt:
            log.info("Aborted.")
            exit(1)


def confirm_edge_creation():
    if config["yes"]:
        log.debug("The option --yes is set. Skipping confirmation...")
        return

    while True:
        try:
            log.info("Press ENTER to continue and create the edges.")
            log.info("Press ctrl+c (or type 'a') to abort. Type 'i' for more info.")

            reply = input().lower()
            if reply != "" and reply != "a" and reply != "i":
                continue

            elif reply == "a":
                log.info("Aborted.")
                exit(1)

            elif reply == "i":
                log.info("")
                log.info('You can set "ask_to_confirm_edge_creation: false" in the config file'
                         " to permanently disable this question.")
                log.info("Add '-y' to the command line to skip this question in the future.")
                log.info("")

            else:
                return
        except KeyboardInterrupt:
            log.info("Aborted.")
            exit(1)


def wait_for_socket(ip, port, timeout, throw=False):
    @tenacity.retry(
        stop=tenacity.stop_after_delay(timeout),
        retry=tenacity.retry_if_exception_type(socket.error),
        wait=tenacity.wait_fixed(1),
        before_sleep=tenacity.before_sleep_log(log, logging.DEBUG),
        reraise=True,
    )
    def try_to_connect():
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.settimeout(1)
            sock.connect((ip, port))
        finally:
            sock.close()

    try:
        log.debug(f"Waiting for connection to {ip}:{port} to be opened, waiting for {timeout} seconds")
        try_to_connect()
        log.debug(f"Connection to {ip}:{port} was opened successfully")
        return True
    except socket.timeout as e:
        if throw:
            raise
        log.warning(str(e))
        return False


def to_yaml(value):
    """
    Convert given value (usually a dict) to YAML, correctly handling multi-line strings.

    :param val:   Object to dump as YAML
    :return:      YAML string
    """
    # for some reason the yaml.dump() below doesn't play well
    # with dict's descendant classes
    if isinstance(value, dict) and type(value) is not dict:
        value = dict(value)

    class MultiLineStringDumper(yaml.Dumper):
        def __init__(self, stream, **kwargs):
            super(self.__class__, self).__init__(stream, **kwargs)
            self.__class__.add_representer(str, self._str_presenter)

        def ignore_aliases(self, data):
            # We don't want yaml aliases, they could be confusing to some people. (CM-12080)
            return True

        @staticmethod
        def _str_presenter(dumper, data):
            if len(data.splitlines()) > 1:  # check for multiline string
                return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
            return dumper.represent_scalar("tag:yaml.org,2002:str", data)

    return yaml.dump(value, default_flow_style=False, Dumper=MultiLineStringDumper)


def cod_log(logger, action, percentage):
    logger.info("## Progress: %s", percentage)
    logger.info("#### stage: %s""", action)


def get_package_type(bcm_distro):
    if bcm_distro.upper().startswith("CENTOS"):
        return "yum"
    if bcm_distro.upper().startswith("ROCKY"):
        return "yum"
    elif bcm_distro.upper().startswith("ALMA"):
        return "yum"
    elif bcm_distro.upper().startswith("SLES"):
        return "zypper"
    elif bcm_distro.upper().startswith("SL"):  # This has to come after SLES
        return "yum"
    elif bcm_distro.upper().startswith("RHEL"):
        return "yum"
    elif bcm_distro.upper().startswith("UBUNTU"):
        return "apt"
    else:
        raise KeyError("Unknown distro %s" % bcm_distro)


def enumerate_repo_config_files(bcm_distro, root_dir="/", suffix="", sles_suffix="", prefix_root_dir=True):
    if not sles_suffix:
        sles_suffix = suffix
    os_repo_config_mapping = {
        "yum": [
            "etc/yum.repos.d/cm{}.repo".format(suffix)
        ],
        "zypper": [
            "etc/zypp/repos.d/Cluster_Manager{}_Updates.repo".format(sles_suffix)
        ],
        "apt": [
            "etc/apt/sources.list.d/cm{}.list".format(suffix)
        ]
    }

    if prefix_root_dir:
        prefix_dir = root_dir
    else:
        prefix_dir = "/"

    software_images = [
        os.path.join(prefix_dir, "cm/images", elem) for elem in
        os.listdir(os.path.join(root_dir, "cm/images"))
    ]
    node_installer = [os.path.join(prefix_dir, "cm/node-installer")]
    root_paths = [prefix_dir] + software_images + node_installer

    package_type = get_package_type(bcm_distro)
    repo_files = os_repo_config_mapping[package_type]

    for root_path in root_paths:
        for repo_file in repo_files:
            config_file = os.path.join(root_path, repo_file)
            yield config_file


@static_vars(escape_slash_re=re.compile(r"([^\\])\/"))
def sed_file_inplace(path, old, new, plain_text=False):
    def escape_all_slashes(s):
        while True:
            s, count = sed_file_inplace.escape_slash_re.subn(r"\1\/", s)
            if count == 0:
                return s

    if plain_text:
        old = re.escape(old)
        new = re.escape(new)

    try:
        old = escape_all_slashes(old)
        new = escape_all_slashes(new)
        expression = "s/%s/%s/g" % (old, new)
        check_call(["sed", "-i", "-E", expression, path])
    except CalledProcessError as e:
        log.warning(str(e))
        raise


def _define_dev_repository(repo, repo_type, bcm_distro, bcm_version):

    major_version, _ = _extract_bcm_version(bcm_version)

    if bcm_distro.lower().startswith("ubuntu"):

        # The regex below is expecting it in lowercase
        bcm_distro = bcm_distro.lower()
        # The regex below is expecting ubuntu1604 instead of ubuntu16.04
        bcm_distro = bcm_distro.replace(".", "")
        # Retrieve distro/version path from distro string (e.g., ubuntu1604 -> ubuntu/1604)
        ubuntu_version = re.match(r"([a-z]+)([0-9]+)", bcm_distro, re.I).groups()[1]

        # Comment out all updates-*.brightcomputing.com repo lines
        sed_file_inplace(
            repo,
            "^(deb http://([^@]+@)?updates[^.]*.brightcomputing.com)",
            "# \\1"
        )
        dev_repo_line = string.Template(UBUNTU_DEV_REPO).substitute(
            repo_type=repo_type,
            bcm_version=major_version,
            ubuntu_version=ubuntu_version,
        )
        with open(repo, "r+") as repo_file:
            if not any([line.strip() == dev_repo_line for line in repo_file.readlines()]):
                repo_file.write("\n" + dev_repo_line)
    else:
        sed_file_inplace(repo, major_version, bcm_version, plain_text=True)


def _extract_bcm_version(bcm_version):
    pattern = r"(.*(?=\-))(.*)"

    try:
        return re.match(pattern, bcm_version).groups()
    except AttributeError:
        log.error(
            "Failed to extract version from BCM version '%s'" % bcm_version
        )
        raise


def define_dev_repositories(bcm_distro, bcm_version, root_dir):
    try:
        cm_repos = enumerate_repo_config_files(bcm_distro, root_dir)
        ml_repos = enumerate_repo_config_files(bcm_distro, root_dir, suffix="-ml", sles_suffix="_ML")
        ni_repos = enumerate_repo_config_files(bcm_distro, root_dir, suffix="-ni", sles_suffix="_NI")
    except KeyError:
        log.warning(
            "Not changing repository configuration files for unknown distro '%s'" % (bcm_distro, )
        )

    for repo, repo_type in itertools.chain(
        zip(cm_repos, itertools.repeat("cm")),
        zip(ml_repos, itertools.repeat("ml")),
        zip(ni_repos, itertools.repeat("ni"))
    ):
        if not os.path.isfile(repo):
            log.debug("Skipping '%s', does not exist" % (repo, ))
            continue
        log.debug("Enabling %s repository in '%s'" % (bcm_version, repo))
        _define_dev_repository(repo, repo_type, bcm_distro, bcm_version)


def get_user_at_fqdn_hostname():
    return "{user}@{host}".format(
        user=getpass.getuser(),
        host=socket.getfqdn()
    )


class MultithreadRunError():
    """Used by multithread_run to indicate error."""
    def __init__(self, traceback, exception):
        self.traceback = traceback
        self.exception = exception


def multithread_run(function, args_list, max_threads):
    """Runs a function multiple times in different threads and waits for them.

    For each tuple in 'args_list', a new thread will be created to run function(*args_list[i]) with a maximum
    of 'max_threads' in parallel.

    The most common use case is to run the same function with different arguments. But It's possible to run
    different functions by passing them in the 'args_list' and setting a dummy function in 'function':
    multithread_run(lambda f: f(), [func1, func2])
    """

    def wrapped(args):
        try:
            if not isinstance(args, tuple):
                args = (args,)
            return function(*args)
        except Exception as e:
            return MultithreadRunError(traceback.format_exc(), e)

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
        futures = [executor.submit(wrapped, args) for args in args_list]
        _, not_done = concurrent.futures.wait(futures, timeout=600)
        while not_done:
            try:
                _, not_done = concurrent.futures.wait(futures, timeout=600)
            except concurrent.futures.TimeoutError:
                log.debug("Timed out waiting for asynchronous result. Waiting again...")
                continue

    return [x.result() for x in futures]


def running_in_virtual_env():
    if "VIRTUAL_ENV" in os.environ:
        return True
    return False


def partition(func, iterable):
    true_items = []
    false_items = []
    for item in iterable:
        if func(item):
            true_items.append(item)
        else:
            false_items.append(item)
    return true_items, false_items


def resolve_hostname(hostname: str, retries=4, retry_delay_sec=0.2) -> str:
    ip = None
    while True:
        try:
            ip = socket.gethostbyname(hostname)
            log.debug(f"Successfully resolved '{hostname}' to '{ip}'.")
            break
        except socket.error as e:
            log.debug(f"Failed to resolve hostname '{hostname}' to an IP. Reason: {e}")
            if retries > 0:
                log.debug(f"... Retry in {retry_delay_sec} seconds.")
                retries -= 1
                time.sleep(retry_delay_sec)
            else:
                raise CODException(f"Failed to resolve hostname '{hostname}' to an IP address.")
    return ip


def resolve_hostnames(hostnames: List[str]) -> Dict[str, str]:
    return {name: resolve_hostname(name) for name in hostnames}


def get_commands_to_inject_resolved_hostnames(roots: List[str], resolved_hostname_ips: Dict[str, str]) -> List[str]:
    commands = []

    for root in roots:
        if root == "/":
            root = ""
        commands += [f"echo '# This section was added by cod-os' >> {root}/etc/hosts"]
        for hostname, ip in resolved_hostname_ips.items():
            commands += [f"echo '{ip}                     {hostname}' >> {root}/etc/hosts"]
        commands += [f"echo '# The end of entries added by cod-os' >> {root}/etc/hosts"]
    return commands


def log_no_clusters_found(command_name):
    log.info(f"There were no eligible clusters to perform the '{command_name}' operation" +
             (f" with the arguments: {', '.join(config['filters'])}"
              if config.is_item_set_explicitly("filters") else ""))


def generate_timedelta(string):
    regex = re.compile(r"((?P<days>\d+?)d)?((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?")
    match = regex.match(string.replace(" ", "")).groupdict()

    days = int(match.get("days") or 0)
    hours = int(match.get("hours") or 0)
    minutes = int(match.get("minutes") or 0)
    delta = timedelta(days=days, hours=hours, minutes=minutes)

    return delta


def generate_older_than_date(delta):
    if not delta:
        return None

    return (datetime.utcnow() - delta).replace(tzinfo=tzutc())


def report_progress(status, previous_status=""):
    """
    Report progress only if the progress is printed to a terminal.
    """
    if not sys.stderr.isatty:
        return ""

    if status != previous_status:
        sys.stderr.write("\r" + status)

    return status


def parse_assignment(s):
    """
    :param s: a string representing an assignment, of the form key=value
    :return: a pair of k, v. k is a non-None string, v is allowed to be empty.
    """
    k, v = s.split("=", 1)  # allows '=' to show up in the value
    assert k, f"The string '{s}' is not a valid assignment in the form 'a=b'"
    return k, v



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import os

from clusterondemand.cidr import cidr
from clusterondemand.clusternameprefix import must_start_with_cod_prefix
from clusterondemand.configuration_validation import (
    generate_password_if_not_set,
    validate_access_to_cluster,
    validate_password_strength
)
from clusterondemand.inbound_traffic_rule import InboundTrafficRule
from clusterondemand.ip import ip, ip_must_be_within_cidr, nth_ip_in_default_network
from clusterondemandconfig import (
    ConfigNamespace,
    if_set_file_exists_and_readable,
    match_http_proxy_format,
    may_not_equal_none,
    require_at_least_one_to_be_set,
    requires_other_parameter_to_be_set
)

from . import localpath
from .clusternameprefix import clusterprefix_ns
from .paramvalidation import ParamValidator
from .utils import confirm_ns

log = logging.getLogger("cluster-on-demand")

ENFORCING_CONFIG_FILES = [
    "/etc/cm-cluster-on-demand.d/enforced.ini"
]

# in cluster-on-demand/clusterondemand there is another configuration file reserved for CI (ci.ini)
# this file was left out from the INCLUDED_CONFIG_FILES list on purpose, in order not to load it in
# virtual environments. DO NOT INCLUDE IT HERE!!!
INCLUDED_CONFIG_FILES = [
    os.path.join(os.path.dirname(__file__), "krusty.ini"),
    os.path.join(os.path.dirname(__file__), "credentials.ini"),
    os.path.join(os.path.dirname(__file__), "bender.ini"),
]
SYSTEM_CONFIG_FILES = [
    "/etc/cm-cluster-on-demand.ini",
    "/etc/cm-cluster-on-demand.conf",
    "/etc/cm-cluster-on-demand.d/*",
]
USER_CONFIG_FILES = [
    os.path.expanduser("~/cm-cluster-on-demand.ini"),
    os.path.expanduser("~/cm-cluster-on-demand.conf"),
    os.path.expanduser("~/cm-cluster-on-demand.d/*")
]

NO_WLM = "dont-configure"

WLMS_8_2 = [
    NO_WLM,
    "slurm",
    "sge",
    "pbspro",
    "pbspro-ce",
    "openpbs",
    # LSF and UGE are supported by Bright, but not ouf of the box
]

ALL_WLMS = list(set(WLMS_8_2))

configuration_ns = ConfigNamespace("configuration.output", "configuration parameters")
# TODO: move namespace with system_config and config parameters to cod-config.
#  They are used by cod-config and should be defined there.
configuration_ns.add_switch_parameter(
    "show_configuration",
    advanced=True,
    help="Show configuration before actual command execution")
configuration_ns.add_switch_parameter(
    "show_secrets",
    advanced=True,
    help=("When dumping the configuration, it should show the values of secret parameters, "
          "e.g. passwords, keys, etc."))
configuration_ns.add_switch_parameter(
    "check_config",
    advanced=True,
    default=True,
    help="Check configuration doesn't have unknown sections and parameters.")
configuration_ns.add_switch_parameter(
    "check_config_for_all_commands",
    advanced=True,
    default=False,
    help=("Check configuration for all known COD-commands. By default, only config sections"
          " related to current command are checked. This makes config check much slower"))
configuration_ns.add_switch_parameter(
    "system_config",
    advanced=True,
    boot=True,
    default=True,
    help="Load system config files implicitly. If disabled, only config files specified with --config are loaded")
configuration_ns.add_enumeration_parameter(
    "config",
    boot=True,
    default=None,
    flags=["-c"],
    help="Extra config files")

eula_ns = ConfigNamespace("common.eula", help_section="eula parameters")
eula_ns.add_switch_parameter(
    "accept_eula",
    advanced=True,
    help="Accept the most recent EULA and skip the dialog")
eula_ns.add_parameter(
    "user_accepted_eula_file",
    advanced=True,
    default=os.path.expanduser("~/.cm-cod-accepted-eula"),
    help="The file location where the EULA accepted by the user is saved.",
    parser=os.path.expanduser,
    validation=may_not_equal_none)


events_ns = ConfigNamespace("common.events", help_section="event settings")
events_ns.add_switch_parameter(
    "events",
    advanced=True,
    help="Make the cod client send JSON HTTP events to "
         "the URL specified with --events-url . Events report which 'cod' sub "
         "commands are being used and in what way.")
events_ns.add_parameter(
    "events_url",
    advanced=True,
    help="HTTP URL to which the events are sent, e.g. http://example:12345")
events_ns.add_parameter(
    "events_timeout",
    advanced=True,
    default=1,
    help_varname="SECONDS",
    help="Max time in seconds to try to connect and submit the event. Timeouts are not fatal.")
events_ns.add_switch_parameter(
    "events_in_ve",
    advanced=True,
    default=False,
    validation=requires_other_parameter_to_be_set("events"),
    help="Send events to the event URL even when executing inside of a Python virtual environment")


common_ns = ConfigNamespace("common", help_section="common parameters")
common_ns.import_namespace(configuration_ns)
common_ns.import_namespace(eula_ns)
common_ns.import_namespace(events_ns)
common_ns.add_switch_parameter(
    "help",
    flags=["-h"],
    help="Show this message and exit")
common_ns.add_switch_parameter(
    "advanced_help",
    flags=["-H"],
    help="Don't omit advanced configuration parameters from the help output. Implies --help")
common_ns.add_parameter(
    "explain",
    help=(
        "Show detailed information about the immediately following parameter."
        " Which can be a name, a regular expression, ENV VAR or another flag."
    ))
common_ns.add_switch_parameter(
    "verbose",
    flags=["-v", "-vv", "-vvv"],
    help="Verbosity level")
common_ns.add_switch_parameter(
    "print_stack_trace_on_failure",
    default=False,
    help="Print the stack trace on failure, even if -v wasn't specified")
common_ns.add_parameter(
    "log_file",
    advanced=True,
    help="File to write full debug log to",
    parser=os.path.expanduser)
common_ns.add_switch_parameter(
    "log_omit_time",
    advanced=True,
    help="Omit timestamp information in console log")
common_ns.add_switch_parameter(
    "pdb_on_error",
    advanced=True,
    help="Run PDB post mortem on error")
common_ns.add_switch_parameter(
    "only_ascii_arguments",
    default=True,
    advanced=True,
    help="Only allow for ASCII characters to be used in CLI arguments. "
         "Useful when pasting values from user interfaces which might contain invisible non-ASCII "
         "characters.")
common_ns.add_parameter(
    "max_threads",
    default=10,
    advanced=True,
    help="Maximum number of threads for parallel operations."
)
common_ns.add_switch_parameter(
    "version",
    help="Displays version information about cluster-on-demand.",
)


clustercreatelicense_ns = ConfigNamespace(
    "cluster.create.license",
    help_section="Bright Cluster Manager licensing information")
clustercreatelicense_ns.add_parameter(
    "license_unit",
    help="License unit")
clustercreatelicense_ns.add_parameter(
    "license_locality",
    help="License locality")
clustercreatelicense_ns.add_parameter(
    "license_country",
    default="US",
    help="Two characters",
    validation=may_not_equal_none)
clustercreatelicense_ns.add_parameter(
    "license_product_key",
    env="LICENSE_PRODUCT_KEY",
    help="Bright Cluster Manager Product Key",
    secret=True,
    validation=[may_not_equal_none, ParamValidator.validate_license_product_key])
clustercreatelicense_ns.add_parameter(
    "license_organization",
    help="Name of your organization")
clustercreatelicense_ns.add_parameter(
    "license_state",
    help="Name of your state or province")
clustercreatelicense_ns.add_parameter(
    "license_activation_token",
    env="LICENSE_ACTIVATION_TOKEN",
    advanced=True,
    help="Bright Cluster Manager license activation token. It can be used instead of a product "
         "key. If both are specified, activation token will take precedence (the product key, "
         "if also specified, will be ignored).",
    secret=True)
clustercreatelicense_ns.add_parameter(
    "license_activation_url",
    env="LICENSE_ACTIVATION_URL",
    advanced=True,
    help="Bright Cluster Manager license activation URL. "
         "It should point to a server which is capable of issuing Bright licences after being given "
         "the License Activation Token as input. "
         "Only makes sense if license_activation_token is specified.")
clustercreatelicense_ns.add_parameter(
    "license_activation_password",
    env="LICENSE_ACTIVATION_PASSWORD",
    advanced=True,
    help="Bright Cluster Manager license activation password. "
         "It will be used by cm-bright-setup for client authentication with the license activation server.",
    secret=True)
clustercreatelicense_ns.add_parameter(
    "http_proxy",
    advanced=True,
    default=None,
    help="HTTP proxy to be used on the headnode for contacting the bright licensing server.",
    validation=match_http_proxy_format
)


clustercreatepassword_ns = ConfigNamespace(
    "cluster.create.password",
    help_section="root login method to the head node")
clustercreatepassword_ns.add_switch_parameter(
    "log_cluster_password",
    help="Log cluster password to the screen and log files. This option is mandatory if no custom "
         "password, nor SSH keypairs, were specified.")
clustercreatepassword_ns.add_parameter(
    "cluster_password",
    help="The root user password to the cluster. If not specified, a random one will be generated "
         "(use --log-cluster-password to see it). This is also the root user SQL password on the "
         "head node. Upon cluster creation the password is stored in the "
         "/cm/local/apps/cmd/etc/cmd.conf on the head node.",
    secret=True,
    validation=[validate_password_strength, generate_password_if_not_set])
clustercreatepassword_ns.add_parameter(
    "ssh_password_authentication",
    type=bool,
    default=False,
    help="If set to true, it will be possible to SSH to the head node using a "
         "password. This option should NOT be used in untrusted environments as it exposes "
         "the head node to brute force login attacks."),
clustercreatepassword_ns.add_switch_parameter(
    "access_validation",
    default=True,
    help="Causes the cluster creation process to abort early if it won't be able to "
         "guarantee that the SSH access to the cluster will be possible. "
         "Disabling it is useful when e.g. the public SSH key is being delivered to the image in some other "
         "way, than the usual command line argument. "
         "Note, that access validation does not attempt to actually connect to the cluster. "
         "Instead, it merely tries to predict whether the cluster will be accessible to the user, "
         "given the specified argument combination.")
clustercreatepassword_ns.add_parameter(
    "weak_password_policy",
    advanced=True,
    choices=["allow", "warn", "fail"],
    default="warn",
    help="Determine the behavior of cluster create upon the detection of a weak password, when SSH password access "
         "is enabled. When set to 'allow', a weak password will silently be accepted; setting it to 'warn' "
         "triggers printing a warning to the user when a weak password is detected; setting it to 'fail' causes the "
         "cluster creation to abort when a weak password is detected")
clustercreatepassword_ns.add_parameter(
    "ssh_pub_key_path",
    help="Path to the public key",
    help_varname="PATH_TO_FILE",
    parser=localpath.localpath,
    validation=[localpath.must_exist, localpath.must_be_readable])
clustercreatepassword_ns.add_validation(validate_access_to_cluster)

clustercreatewlm_ns = ConfigNamespace("cluster.create.wlm")
clustercreatewlm_ns.add_parameter(
    "wlm",
    choices=ALL_WLMS,
    default=NO_WLM,
    help="Workload Manager of choice. "
         "This can also be configured later. "
         "Note that this list is a subset of the WLM systems supported across different versions of Bright. "
         "Therefore, some WLM systems are only configurable later on, after the cluster has been created. "
         "Defaults to 'dont-configure'. ",
    help_choices={NO_WLM: "do not configure any (can be configured later)"})


clustercreateconfirm_ns = ConfigNamespace("cluster.create.confirm")
clustercreateconfirm_ns.import_namespace(confirm_ns)
clustercreateconfirm_ns.add_parameter(
    "ask_to_confirm_cluster_creation",
    default=True,
    help="Ask for confirmation when creating a new cluster. Use -y to skip this question.")


clustercreatedisks_ns = ConfigNamespace("cluster.create.disks")
clustercreatedisks_ns.add_parameter(
    "head_node_root_volume_size",
    default=50,
    help="Head node root disk size in GB",
    help_varname="SIZE_IN_GB",
    type=int)

nodes_ns = ConfigNamespace("cluster.create.nodes")
nodes_ns.add_parameter(
    "nodes",
    default=5,
    flags=["-n"],
    help="The amount of cloud nodes to configure for a cluster. "
    "The nodes are not powered on automatically.",
    type=int)

clustercreatename_ns = ConfigNamespace("cluster.create.name")
clustercreatename_ns.add_parameter(
    "name",
    help=("(default: auto) Name of the cluster to create. "
          "By default the name is generated from version, distro and creation timestamp."),
    validation=must_start_with_cod_prefix
)

bright_setup_ns = ConfigNamespace("cluster.brightsetup")
bright_setup_ns.add_enumeration_parameter(
    "prebs",
    default=[],
    help_varname="COMMAND",
    help=("Command(s) executed by cloud-init before "
          "cm-bright-setup (before CMDaemon starts). Useful for package update."
          " Multiple arguments are allowed")
)
bright_setup_ns.add_enumeration_parameter(
    "postbs",
    default=[],
    help_varname="COMMAND",
    help="Command(s) executed by cloud-init post cm-bright-setup (Once CMDaemon starts)"
)
bright_setup_ns.add_parameter(
    "run_cm_bright_setup",
    default=True,
    help=("Whether or not to initialize the cluster by running cm-bright-setup "
          "(activate license, etc)"),
)
CFG_NO_ADMIN_EMAIL = "none"
CFG_AUTO_ADMIN_EMAIL = "auto"
bright_setup_ns.add_parameter(
    "admin_email",
    default=CFG_NO_ADMIN_EMAIL,
    help="Admin email address to set in CMDaemon."
)

clustercreate_ns = ConfigNamespace("cluster.create")
clustercreate_ns.import_namespace(clustercreatelicense_ns)
clustercreate_ns.import_namespace(clustercreatepassword_ns)
clustercreate_ns.import_namespace(clustercreatewlm_ns)
clustercreate_ns.import_namespace(clusterprefix_ns)
clustercreate_ns.import_namespace(clustercreateconfirm_ns)
clustercreate_ns.import_namespace(clustercreatedisks_ns)
clustercreate_ns.import_namespace(nodes_ns)
clustercreate_ns.import_namespace(bright_setup_ns)

clustercreate_ns.add_switch_parameter(
    "dry_run",
    flags=["-d"],
    help="Dry run - do not actually create the cluster")
clustercreate_ns.add_parameter(
    "on_error",
    choices=["cleanup", "abort"],
    help="What to do on failure",
    help_choices={"cleanup": "destroy only failed clusters", "abort": "do nothing and exit"})
clustercreate_ns.add_parameter(
    "head_node_type",
    flags=["-m"],
    help="The instance type must exist in the region you use.")
clustercreate_ns.add_parameter(
    "node_type",
    help="The instance type of compute nodes. It must exist in the region you use.")
clustercreate_ns.add_enumeration_parameter(
    "inbound_rule",
    help=("One or several inbound traffic rules for the cluster's head node in the following format: "
          "[src_cidr[:src_port],]dst_port[:protocol]. Where port can be a single port or a dash separated range and "
          "supported protocols are: tcp, udp.  A wildcard value will be assumed for every optional non-provided "
          "parameter (e.g. all ports, all protocols, all IPs)  Examples: '80' '21:udp' '11.0.0.0/24,20-23:TCP' "
          "'12.0.0.0/32:6000-6500,443'"),
    default=[InboundTrafficRule(rule) for rule in ["22:tcp", "80:tcp", "443:tcp", "8081:tcp"]])


clusterdelete_ns = ConfigNamespace("cluster.delete")
clusterdelete_ns.import_namespace(confirm_ns)
clusterdelete_ns.add_positional_parameter(
    "name",
    help="Name of the cluster to delete",
    require_value=True)


clusterlist_ns = ConfigNamespace("cluster.list")
clusterlist_ns.add_parameter(
    "output_format",
    flags=["-f"],
    default="table",
    choices=["table", "csv", "value", "yaml"],
    help="Output format for lists")

sshconfig_ns = ConfigNamespace("sshconfig", "SSH local configuration")
sshconfig_ns.add_parameter(
    "ssh_config_path",
    help="Path to local ssh config to save COD cluster ssh configuration in.",
    default="~/.ssh/config",
)
sshconfig_ns.add_switch_parameter(
    "update_ssh_config",
    help="Updates the contents of ~/.ssh/config, this is used to maintain the local ssh configuration "
         "access with ssh, scp and related tools.",
)
sshconfig_ns.add_parameter(
    "ssh_config_alias_prefix",
    help="Prefix to be used when populating host entries for the cluster head nodes in the COD "
         "section of ~/.ssh/config. Default is ''.",
    default=""
)


cmd_debug_ns = ConfigNamespace("cmddebug", "Debug settings for CMDaemon")
cmd_debug_ns.add_switch_parameter(
    "cmd_debug",
    advanced=True,
    help="Enable debug mode on CMDaemon. The debug mode is enabled before CMDaemon is started for the first time. "
         "The affected subsystems can be customized with --cmd-debug-subsystems."
)
cmd_debug_ns.add_parameter(
    "cmd_debug_subsystems",
    default="*",
    help_varname="SUBSYSTEMS",
    advanced=True,
    help="Which CMDaemon subsystems (e.g: PROV, CMD, SERVICE) should have debug logging enabled. If --cmd-debug is "
         "set, SUBSYSTEMS is written to /cm/local/apps/cmd/etc/logging.cmd.conf on the head node and the default "
         "software image. Defaults to '*', i.e. all subsystems."
)

append_to_bashrc_ns = ConfigNamespace("cluster.create.headnode")
append_to_bashrc_ns.add_enumeration_parameter(
    "append_to_root_bashrc",
    default=[],
    help_varname="ENV=VAR",
    help="Lines to append to the /root/.bashrc file on the head node"
)

aws_provider_tag_cm_setup_ns = ConfigNamespace("cluster.create.headnode")
aws_provider_tag_cm_setup_ns.add_parameter(
    "inject_cod_prefix_as_aws_provider_tag",
    advanced=True,
    default=False,
    help=("If set to true, COD_PREFIX=<prefix_value> will be added to list of AWS tags "
          "in /root/cm-setup.conf on the new cluster, and that key=value pair will be used "
          "to tag AWS VMs. This can be used to track AWS expenses "
          "of multiple COD users sharing the same AWS account."),
    help_varname="true|false"
)

internal_net_ns = ConfigNamespace("cluster.internalnet")
internal_net_ns.add_parameter(
    "internal_cidr",
    default=cidr("10.141.0.0/16"),
    help="CIDR of the cluster's internal network.",
    parser=cidr,
)
internal_net_ns.add_parameter(
    "head_node_internal_ip",
    advanced=True,
    default=nth_ip_in_default_network(-2, "internal_cidr"),
    help_varname="IP",
    help=("IP address of the head node on internal network. "
          "If not set will be calculated from internal_cidr."),
    parser=ip,
    validation=ip_must_be_within_cidr("internal_cidr"),
)
internal_net_ns.add_parameter(
    "internal_mtu",
    type=int,
    help="MTU of the cluster's internal network. "
)

node_disk_setup_ns = ConfigNamespace("cluster.node")
node_disk_setup_ns.add_validation(require_at_least_one_to_be_set("node_disk_setup_path", "node_disk_setup"))
node_disk_setup_ns.add_parameter(
    "node_disk_setup_path",
    help_varname="PATH_TO_XML",
    help="Path to the XML file with the disk setup which is to be used for the nodes",
    validation=if_set_file_exists_and_readable,
)
node_disk_setup_ns.add_parameter(
    "node_disk_setup",
    default="""
<?xml version='1.0' encoding='ISO-8859-1'?>
<!-- This is the CoD default -->
<!-- Just a single xfs partition -->
<diskSetup xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance'>
  <device>
    <blockdev>/dev/vdb</blockdev>
    <blockdev>/dev/sda</blockdev>
    <blockdev mode='cloud'>/dev/sdb</blockdev>
    <blockdev mode='cloud'>/dev/hdb</blockdev>
    <blockdev mode='cloud'>/dev/vdb</blockdev>
    <blockdev mode='cloud'>/dev/xvdb</blockdev>
    <blockdev mode='cloud'>/dev/xvdf</blockdev>
    <blockdev mode="cloud">/dev/nvme1n1</blockdev>
    <partition id='a2'>
      <size>max</size>
      <type>linux</type>
      <filesystem>xfs</filesystem>
      <mountPoint>/</mountPoint>
      <mountOptions>defaults,noatime,nodiratime</mountOptions>
    </partition>
  </device>
</diskSetup>
    """)


resolve_hostnames_ns = ConfigNamespace("cluster.create.headnode")
resolve_hostnames_ns.add_enumeration_parameter(
    "resolve_hostnames",
    default=[],
    help_varname="DOMAIN",
    advanced=True,
    help=("Hostname(s) which will be resolved by the tool during the cluster creation process. "
          "Resolved hostnames will be appended to /etc/hosts on the head node(s) and in the software image. "
          "This is useful when working around problems with DNS.")
)

timezone_ns = ConfigNamespace("cluster.create")
timezone_ns.add_parameter(
    "timezone",
    default="Europe/Amsterdam",
    help="Timezone of the cluster",
    validation=may_not_equal_none
)


def get_system_config_files(additional_included_config_files=None, additional_system_config_files=None):
    return INCLUDED_CONFIG_FILES + \
        (additional_included_config_files or []) + \
        SYSTEM_CONFIG_FILES + \
        (additional_system_config_files or []) + \
        USER_CONFIG_FILES


def get_enforcing_config_files():
    return ENFORCING_CONFIG_FILES



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import importlib

import six


def import_if_exists(module):
    """Try to import a module. If the module does not exist, nothing is raised."""
    try:
        return importlib.import_module(module)
    except ImportError as error:
        possible_error_messages = ["No module named %s" % (fragment) for fragment in _module_path_fragments(module)]

        if str(error) in possible_error_messages:
            return None
        else:
            raise


def load_module_attribute(module_path, attribute):
    """Return the value of `module_path`.`attribute`, None if the module is not available."""
    module = import_if_exists(module_path)
    return eval("module.%s" % (attribute)) if module else None


if six.PY2:
    def _module_path_fragments(module_path):
        """Convert a module path like 'foo.bar.baz' to ['foo.bar.baz', 'bar.baz', 'baz']."""
        fragments = []
        while "." in module_path:
            fragments.append(module_path)
            module_path = module_path[module_path.index(".") + 1:]
        return fragments + [module_path]
elif six.PY3:
    def _module_path_fragments(module_path):
        """Convert a module path like 'foo.bar.baz' to ["'foo'", "'foo.bar'", "'foo.bar.baz'"]."""
        fragments = []
        while "." in module_path:
            fragments.append("'%s'" % (module_path))
            module_path = module_path[:module_path.index(".")]
        return fragments + ["'%s'" % (module_path)]



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

from itertools import chain

from clusterondemandconfig import ConfigNamespace, config, human_readable_config_dump, load_configuration_for_parameters

from . import configuration
from .import_utils import load_module_attribute

config_ns = ConfigNamespace("config.show")
config_ns.import_namespace(configuration.common_ns)


COMMAND_HELP_TEXT = """
Display the entire configuration within a table. Note that this command takes into account any existing
config files, and environment variables, both of which which might impact the values in the generated config;
to avoid that, use '--no-system-config', which will generate a config from only the hardcoded defaults and
the config files specified with --config.
"""


def _load_module_parameters(module_name, commands_name):
    commands = load_module_attribute(module_name, commands_name)
    return [] if commands is None else commands.parameters()


def run_command():
    aws_parameters = _load_module_parameters("clusterondemandaws.cli", "aws_commands")
    azure_parameters = _load_module_parameters("clusterondemandazure.cli", "azure_commands")
    openstack_parameters = _load_module_parameters("clusterondemandopenstack.cli", "openstack_commands")
    vmware_parameters = _load_module_parameters("clusterondemandvmware.cli", "vmware_commands")

    config_to_dump = load_configuration_for_parameters(
        chain(aws_parameters, azure_parameters, openstack_parameters, vmware_parameters),
        system_config_files=configuration.SYSTEM_CONFIG_FILES,
        enforcing_config_files=configuration.ENFORCING_CONFIG_FILES
    )

    print(human_readable_config_dump(config_to_dump, show_secrets=config["show_secrets"]))



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from clusterondemand.bcm_version import BcmVersion
from clusterondemandconfig import config

log = logging.getLogger("cluster-on-demand")


def get_license_dict(name=None):
    license_dict = {
        "cluster_name": str(name if name else config["name"]),
        "unit": str(config["license_unit"]),
        "locality": str(config["license_locality"]),
        "country": str(config["license_country"]),
        "product_key": str(config["license_product_key"]),
        "state": str(config["license_state"]),
        "organization": str(config["license_organization"])
    }
    # COD images older than 8.1-11 will break because of unrecognized "license_activation_token"
    # configuration key. To prevent it, don't pass it unless it was specified.
    if config["license_activation_token"]:
        license_dict["activation_token"] = str(config["license_activation_token"])
        if config["license_activation_url"]:
            license_dict["activation_url"] = str(config["license_activation_url"])
    return license_dict


def generate_bright_setup(cloud_type,
                          wlm,
                          license_dict,
                          hostname,
                          password,
                          node_count,
                          timezone,
                          admin_email=None,
                          node_disk_setup_path=None,
                          node_kernel_modules=None):
    brightsetup = {
        "cloud_type": cloud_type,
        "bright": {
            "wlm": wlm,
            "wlm_slot_count": "AUTO" if BcmVersion(config["version"]) > "9.1" else 1,
            # None   type gets translated to 'None' string somewhere,
            # that's why need need empty string here for pbsproc_lic_server
            "pbspro_lic_server": "",
            "license": license_dict,
            "hostname": str(hostname),  # otherwise it will end up as unicode:  u"xxxx"
            "master_compute_node": False,
            "password": password,
            "node_count": node_count
        },
    }

    if node_kernel_modules:
        brightsetup["bright"]["node_kernel_modules"] = node_kernel_modules
    else:
        # TODO:fixme: CM-9026 otherwise somehow the default config value does not kick in
        # (we get key error)
        brightsetup["bright"]["node_kernel_modules"] = []
    if node_disk_setup_path:
        brightsetup["bright"]["node_disk_setup_path"] = node_disk_setup_path
    else:
        # TODO:fixme: CM-9026 otherwise somehow the default config value does not kick in
        # (we get key error)
        brightsetup["bright"]["node_disk_setup_path"] = ""
    if timezone:
        brightsetup["bright"]["timezone"] = timezone
    if admin_email:
        brightsetup["bright"]["admin_email"] = admin_email

    if BcmVersion(config["version"]) > "8.1" and config["http_proxy"] is not None:
        brightsetup["bright"]["http_proxy"] = config["http_proxy"]
    elif config["http_proxy"] is not None:
        log.warning("http_proxy is not supported for versions prior to 8.2.")

    return {
        "modules": {
            "brightsetup": brightsetup
        }
    }



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import netaddr

from .exceptions import CODException


def ip(string):
    """A parser for ip strings."""
    try:
        ip_addr = netaddr.IPAddress(string)
        return ip_addr
    except netaddr.core.AddrFormatError:
        raise CODException("'%s' is not a valid IP" % (string))


def ip_must_be_within_cidr(cidr_name):
    """Validation that raises an error when the IP value of the parameter is not contained within the CIDR"""
    def wrapper(parameter, configuration):
        cidr_cfg = configuration.get_item_for_key(cidr_name).parameter
        ip_addr = configuration[parameter.key]
        cidr = configuration[cidr_cfg.key]

        if ip_addr is None:
            raise CODException("%s is not set" % (parameter.key))
        if cidr is None:
            raise CODException("%s is not set" % (cidr_cfg.key))

        if netaddr.IPAddress(ip_addr) not in netaddr.IPNetwork(cidr):
            raise CODException(
                "%s=%s not within %s=%s. Use %s and/or %s to change that." % (
                    parameter.key, ip_addr,
                    cidr_cfg.key, cidr,
                    cidr_cfg.default_flag, parameter.default_flag
                )
            )

    return wrapper


def nth_ip_in_default_network(nth, network):
    """Generate a dynamic default method for obtaining the nth ip address of a cidr parameter."""
    return lambda _, config: config[network][nth]



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import clusterondemandconfig

from . import configdump, configshow, tabcompletion_generate
from .command_runner import run_invoked_command

cod_commands = clusterondemandconfig.CommandContext("cm-cod")
cod_commands.add_group("cm-cod config", "Configuration operations")
cod_commands.add_command(
    "cm-cod config show",
    configshow,
    configdump.COMMAND_HELP_TEXT,
    require_eula=False
)
cod_commands.add_command(
    "cm-cod config dump",
    configdump,
    configdump.COMMAND_HELP_TEXT,
    require_eula=False
)

cod_commands.add_group("cm-cod tabcompletion", "Tab completion utilities")
cod_commands.add_command(
    "cm-cod tabcompletion generate",
    tabcompletion_generate,
    tabcompletion_generate.COMMAND_HELP_TEXT,
    require_eula=False
)


def cli_main():
    run_invoked_command(cod_commands)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# This file contains internal functions that are useful in test code
#
# Some useful functions are:
#  - patch_clientutils:
#    Decorator to be applied to test functions. This will patch the openstack
#    client libraries to return some default values

from __future__ import absolute_import

import os
import sys
import uuid
from contextlib import contextmanager
from tempfile import NamedTemporaryFile

import heatclient
import six
from six import StringIO
from six.moves import filter
from six.moves.mock import Mock, PropertyMock, patch

from clusterondemand import const
from clusterondemand.command_runner import sys_argv
from clusterondemandconfig import (
    BCM_VERSION,
    global_configuration,
    load_configuration_for_command,
    validate_configuration
)
from clusterondemandconfig.determine_invoked_command import determine_invoked_command


def ipxe_plain_eth0():
    m = Mock(
        spec="",
        status=u"active",
        tags=[],
        container_format=u"bare",
        min_ram=0,
        updated_at=u"2018-03-17T21:46:33",
        visibility=u"public",
        owner=u"3b9bd8793720479f8bb341edc808a499",
        file=u"/v2/images/bba054ed-3ac7-4fcd-9401-ca115cae2eba/file",
        min_disk=0,
        virtual_size=None,
        id=u"bba054ed-3ac7-4fcd-9401-ca115cae2eba",
        size=8388608,
        checksum=u"3661bd33484e39c4e9d75d15fa61b0ea",
        created_at=u"2018-03-17T21:46:32",
        disk_format=u"raw",
        protected=False,
        direct_url=u"rbd://yadayada",
        schema=u"/v2/schemas/image"
    )
    # Weird trick because Mock uses the name attribute
    type(m).name = PropertyMock(return_value="iPXE-plain-eth0")
    return m


def default_boot_image():
    m = Mock(
        spec="",
        status=u"active",
        tags=[],
        container_format=u"bare",
        min_ram=0,
        updated_at=u"2018-03-17T21:46:33",
        visibility=u"public",
        owner=u"3b9bd8793720479f8bb341edc808a499",
        file=u"/v2/images/bba054ed-3ac7-4fcd-9401-ca233cae2eba/file",
        min_disk=0,
        virtual_size=None,
        id=u"bba054ed-3ac7-4fcd-9401-ca233cae2eba",
        size=8388608,
        checksum=u"3661bd33484e39c4e9d75d15fa61b0ea",
        created_at=u"2018-03-17T21:46:32",
        disk_format=u"raw",
        protected=False,
        direct_url=u"rbd://yadayada",
        schema=u"/v2/schemas/image"
    )
    type(m).name = PropertyMock(return_value="default-boot-image")
    return m


def custom_boot_image():
    m = Mock(
        spec="",
        status=u"active",
        tags=[],
        container_format=u"bare",
        min_ram=0,
        updated_at=u"2018-03-17T21:46:33",
        visibility=u"public",
        owner=u"3b9bd8793720479f8bb341edc808a499",
        file=u"/v2/images/bba054ed-3ac7-4fcd-9401-ca115cae2bbb/file",
        min_disk=0,
        virtual_size=None,
        id=u"bba054ed-3ac7-4fcd-9401-ca115cae2bbb",
        size=8388608,
        checksum=u"3661bd33484e39c4e9d75d15fa61b0ea",
        created_at=u"2018-03-17T21:46:32",
        disk_format=u"raw",
        protected=False,
        direct_url=u"rbd://yadayada",
        schema=u"/v2/schemas/image"
    )
    type(m).name = PropertyMock(return_value="custom-boot-image")
    return m


def flavor_boot_image():
    m = Mock(
        spec="",
        status=u"active",
        tags=[],
        container_format=u"bare",
        min_ram=0,
        updated_at=u"2018-03-17T21:46:33",
        visibility=u"public",
        owner=u"3b9bd8793720479f8bb341edc808a499",
        file=u"/v2/images/bba054ed-3ac7-4fcd-9401-ca115cae2baa/file",
        min_disk=0,
        virtual_size=None,
        id=u"bba054ed-3ac7-4fcd-9401-ca115cae2baa",
        size=8388608,
        checksum=u"3661bd33484e39c4e9d75d15fa61b0ea",
        created_at=u"2018-03-17T21:46:32",
        disk_format=u"raw",
        protected=False,
        direct_url=u"rbd://yadayada",
        schema=u"/v2/schemas/image"
    )
    type(m).name = PropertyMock(return_value="flavor-boot-image")
    return m


def bcmh_centos7u4():
    m = Mock(
        spec="",
        status=u"active",
        bcm_image_revision=u"278",
        tags=[u"MesosFailed",
              u"AzureExtensionOK",
              u"OpenStackExtensionOK",
              u"BasicTestsOK",
              u"AWSExtensionOK",
              u"OpenStackTestsFailed"],
        container_format=u"bare",
        min_ram=0,
        updated_at=u"2018-04-10T15:00:49",
        bcm_api_hash=u"e81c77691ab1cfc4bf7baffe7cd8079c",
        owner=u"962ab1fc6e4f4faea89dcfcb796d5d35",
        visibility=u"public",
        file=u"/v2/images/4c75d9eb-62c9-40b0-865c-eeba52d44b39/file",
        min_disk=0,
        bcm_image_type=u"headnode",
        virtual_size=None,
        id=u"4c75d9eb-62c9-40b0-865c-eeba52d44b39",
        bcm_distro=u"centos7u4",
        size=14289993728,
        bcm_cmd_revision=u"132465",
        bcm_image_id=f"centos7u4-{BCM_VERSION}",
        bcm_pkg_hashes="cmdaemon=af65b9f1435,cm-setup=67557b1d0d2,cluster-tools=ce3b53adad",
        bcm_cloud_type=u"openstack",
        bcm_created_at=u"2018-04-08T23:41:06",
        created_at=u"2018-04-08T23:41:06",
        disk_format=u"raw",
        protected=False,
        bcm_package_groups=u"",
        checksum=u"75f90cec8caf62f4fa1ba1111fa0c9f1",
        direct_url=u"rbd://yadayada",
        bcm_version=BCM_VERSION,
        schema=u"/v2/schemas/image"
    )
    type(m).name = PropertyMock(return_value=u"bcmh-centos7u4")
    return m


def ipxe_caas():
    m = Mock(
        spec="",
        status=u"active",
        tags=[],
        container_format=u"bare",
        min_ram=0,
        min_disk=0,
        visibility=u"public",
        id="c5c16f40-166c-4015-92c5-cc834534dd67",
        size=8388608,
        disk_format=u"raw",
        protected=False,
        created_at=u"2018-04-08T23:41:06",
        direct_url=u"rbd://yadayada",
        schema=u"/v2/schemas/image"
    )
    # Weird trick because Mock uses the name attribute
    type(m).name = PropertyMock(return_value="ipxe-caas")
    return m


def ipxe_plain_eth1():
    m = Mock(
        spec="",
        status=u"active",
        tags=[],
        container_format=u"bare",
        min_ram=0,
        min_disk=0,
        visibility=u"public",
        id="96db1397-7a50-4db5-b096-9698af282397",
        size=8388608,
        disk_format=u"raw",
        protected=False,
        created_at=u"2018-04-08T23:41:06",
        direct_url=u"rbd://yadayada",
        schema=u"/v2/schemas/image"
    )
    # Weird trick because Mock uses the name attribute
    type(m).name = PropertyMock(return_value="iPXE-plain-eth1")
    return m


FLAVOR_M1_XSMALL = Mock(get_keys=Mock(return_value={}),
                        disk=0,
                        ephemeral=0)
type(FLAVOR_M1_XSMALL).name = PropertyMock(return_value="m1.xsmall")

FLAVOR_M1_MEDIUM = Mock(get_keys=Mock(return_value={}),
                        disk=0,
                        ephemeral=0)
type(FLAVOR_M1_MEDIUM).name = PropertyMock(return_value="m1.medium")

FLAVOR_BAREMETAL = Mock(get_keys=Mock(return_value={
                        "cpu_arch": "x86_64"
                        }),
                        disk=0,
                        ephemeral=0)
type(FLAVOR_BAREMETAL).name = PropertyMock(return_value="baremetal")

FLAVOR_COD_MEDIUM = Mock(get_keys=Mock(return_value={}),
                         ram=4096,
                         vcpus=2,
                         disk=0,
                         ephemeral=0,
                         id="2e631454-bf1b-4f3b-b9c2-c15a11058784")
type(FLAVOR_COD_MEDIUM).name = PropertyMock(return_value="cod.medium")

FLAVOR_COD_XSMALL = Mock(get_keys=Mock(return_value={}),
                         ram=1024,
                         vcpus=1,
                         disk=0,
                         ephemeral=0,
                         id="a39c2730-6988-11e8-adc0-fa7ae01bbebc")
type(FLAVOR_COD_XSMALL).name = PropertyMock(return_value="cod.xsmall")

FLAVOR_COD_CUSTOM_BOOT = Mock(get_keys=Mock(return_value={
                              "node_boot_image": "flavor-boot-image"
                              }),
                              ram=1024,
                              vcpus=1,
                              disk=0,
                              ephemeral=0,
                              id="a39c2730-6999-11e8-adc0-fa7ae01bbebc")
type(FLAVOR_COD_CUSTOM_BOOT).name = PropertyMock(return_value="cod.custom.boot")

TESTNET = {
    "status": "ACTIVE",
    "id": "660e44f0-5037-4231-b9ff-35009d33b041",
    "name": "zz-cod-external"
}

TESTSUBNET = {
    "name": "zz-cod-external-sn",
    "id": "4dc39dda-68ba-11e8-adc0-fa7ae01bbebc",
    "cidr": "192.168.200.0/24",
    "ip_version": 4,
    "network_id": TESTNET["id"],
    "gateway_ip": "192.168.200.254",
    "enable_dhcp": True,
    "allocation_pools": [{"start": "192.168.200.128", "end": "192.168.200.253"}],
    "dns_nameservers": ["8.8.8.8"],
}

TESTROUTER = {
    "name": "zz-cod-router",
    "id": "d01063ee-92ee-426a-8d11-5e792a52b952",
    "external_gateway_info": {
        "network_id": "00010203-0405-0607-0809-0a0b0c0d0e0f",
    },
}

TESTPORT = {
    "device_id": TESTROUTER["id"],
    "fixed_ips": [
        {
            "subnet_id": TESTSUBNET["id"],
            "ip_address": TESTSUBNET["gateway_ip"],
        }
    ]
}


def patch_neutron(fake_networks=None, fake_subnets=None, fake_routers=None, fake_ports=None):
    def decorator_neutron(func):
        @patch("clusterondemandopenstack.clientutils.get_neutron_client")
        def wrapper_neutron(self, get_neutron_client):
            def return_networks_list(name=None):
                if fake_networks is None:
                    return {"networks": []}
                else:
                    return {
                        "networks": [network for network in fake_networks if name == network["name"] or name is None]
                    }

            def return_subnets_list(network_id=None):
                if fake_subnets is None:
                    return {"subnets": []}
                else:
                    return {
                        "subnets": [subnet for subnet in fake_subnets if subnet["network_id"] == network_id or
                                    network_id is None]
                    }

            def return_routers_list(name=None):
                if fake_routers is None:
                    return {"routers": []}
                else:
                    return {
                        "routers": [router for router in fake_routers if name == router["name"] or name is None]
                    }

            def return_ports_list(device_id=None):
                if fake_routers is None:
                    return {"ports": []}
                else:
                    return {
                        "ports": [port for port in fake_ports if device_id == port["device_id"] or port is None]
                    }

            def return_create_network(network):
                return {
                    "network": {
                        "name": network["network"]["name"]
                    }
                }

            def return_create_subnet(subnets):
                return {
                    "subnets": [
                        {
                            "name": subnets["subnets"][0]["name"]
                        }
                    ]
                }

            def return_create_router(router):
                return {
                    "router": {
                        "name": router["router"]["name"],
                        "id": "8d3e0d19-d088-46c0-b6da-9a4164286d75",
                    }
                }

            def return_add_interface_router(router_id, subnet_id):
                return

            get_neutron_client.return_value = Mock(list_networks=return_networks_list,
                                                   list_subnets=return_subnets_list,
                                                   list_routers=return_routers_list,
                                                   list_ports=return_ports_list,
                                                   create_network=return_create_network,
                                                   create_subnet=return_create_subnet,
                                                   create_router=return_create_router,
                                                   add_interface_router=return_add_interface_router,
                                                   )
            func(self)

        return wrapper_neutron
    return decorator_neutron


def patch_clientutils(func):
    """
    Return decorator to be used in a test method.

    This will override the clientutil functions to return Mocked objects.
    The current version is very simple, and will have to be improved in order to test
    more complex things.
    """
    @patch_neutron(fake_networks=[TESTNET], fake_subnets=[TESTSUBNET], fake_routers=[TESTROUTER], fake_ports=[TESTPORT])
    @patch("clusterondemandopenstack.clientutils.get_cinder_client")
    @patch("clusterondemandopenstack.clientutils.get_heat_client")
    @patch("clusterondemandopenstack.clientutils.get_glance_client")
    @patch("clusterondemandopenstack.clientutils.get_nova_client")
    def wrapper(self, get_nova_client, get_glance_client,
                get_heat_client, get_cinder_client, *args, **kwargs):
        def return_images_list(filters=None, page_size=25):
            filters = {} if filters is None else filters
            images_list = [ipxe_plain_eth0(), default_boot_image(), custom_boot_image(), flavor_boot_image(),
                           ipxe_plain_eth1(), bcmh_centos7u4(), ipxe_caas()]

            def filter_func(image):
                for k, v in six.iteritems(filters):
                    if k == "tag":
                        if v[0] not in image.tags:
                            return False
                        else:
                            continue

                    if not hasattr(image, k):
                        return False
                    if v not in getattr(image, k):
                        return False
                return True

            return iter(filter(filter_func, images_list))

        def return_servers_list(**kwargs):
            return []

        def return_flavors_list():
            flavors = [FLAVOR_M1_XSMALL, FLAVOR_M1_MEDIUM, FLAVOR_BAREMETAL,
                       FLAVOR_COD_MEDIUM, FLAVOR_COD_XSMALL, FLAVOR_COD_CUSTOM_BOOT]
            return flavors

        def return_volume_types():
            special_vol = Mock()
            type(special_vol).name = PropertyMock(return_value="special_vol")

            default_vol = Mock()
            type(default_vol).name = PropertyMock(return_value="default")

            return [special_vol, default_vol]

        def return_stacks_get_raise_exception_meaning_lack_of_stack(name):
            raise heatclient.exc.HTTPNotFound(
                "Stack not found (this is a custom "
                "string, should not matter for the tests)")

        get_glance_client.return_value = Mock(images=Mock(list=return_images_list))

        get_heat_client.return_value = \
            Mock(stacks=Mock(get=return_stacks_get_raise_exception_meaning_lack_of_stack))

        get_nova_client.return_value = Mock(servers=Mock(list=return_servers_list),
                                            flavors=Mock(list=return_flavors_list))

        get_cinder_client.return_value = Mock(volume_types=Mock(list=return_volume_types))

        func(self)

    return wrapper


class capture_output(object):
    """
    Capture stdout and make it available as StringIO.

    It can be used as a context manager:
    ```
    with capture_output() as output:
        pass
    ```

    or as a decorator:
    ```
    @capture_output()
    def f(arg1, arg2, output):
        pass
    f(arg1, arg2)
    ```
    """

    def __enter__(self):
        self.old_out, sys.stdout = sys.stdout, StringIO()
        return sys.stdout

    def __exit__(self, exception_type, exception_value, traceback):
        sys.stdout = self.old_out

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, output=sys.stdout, **kwargs)
        return wrapper


class environ(object):
    """Temporarily replace certain mappings of os.environ."""
    def __init__(self, new_mapping):
        self.old_mapping = {
            key: os.environ.get(key) for (key, _) in six.iteritems(new_mapping)
        }
        self.new_mapping = new_mapping

    def __enter__(self):
        for key, value in six.iteritems(self.new_mapping):
            os.environ[key] = value

    def __exit__(self, _exception_type, _exception_value, _traceback):
        for key, value in six.iteritems(self.old_mapping):
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value

    def __call__(self, func):
        def wrapper(_self, *args, **kwargs):
            with self:
                return func(_self, *args, **kwargs)
        return wrapper


def run_command(cli, command_context, validate=False):
    with sys_argv(cli):
        command, clean_sys_argv = determine_invoked_command(command_context)

    with sys_argv(clean_sys_argv):
        config = load_configuration_for_command(command, system_config_files=[], enforcing_config_files=[])

    if validate:
        validate_configuration(config)

    with global_configuration(config):
        command.run_command()


def gen_cod_stack(head_a_ip=None, head_b_ip=None, head_a_uuid=None, head_b_uuid=None, **kwargs):
    """Generate a Mock COD stack object.

    Any argument in kwargs will be used as attributes
    The required fields that were not specified will be set to values to make it a valid cod stack
    """
    stack = Mock()
    if not kwargs.get("stack_name"):
        kwargs["stack_name"] = "clustername"
    if not kwargs.get("id"):
        kwargs["id"] = str(uuid.uuid4())
    if not kwargs.get("tags"):
        kwargs["tags"] = []

    # Append tags to make it a valid cod
    required_tags = [
        const.COD_TAG,
        const.COD_TOOL_VERSION_TAG + "=" + str(const.COD_TOOL_VERSION)
    ]
    for tag in required_tags:
        if tag not in kwargs["tags"]:
            kwargs["tags"].append(tag)

    stack.outputs = []
    if head_a_ip is not None:
        stack.outputs.append({"output_key": "head_a_floating_ip", "output_value": head_a_ip})
    if head_b_ip is not None:
        stack.outputs.append({"output_key": "head_b_floating_ip", "output_value": head_b_ip})
    if head_a_uuid is not None:
        stack.outputs.append({"output_key": "head_a_uuid", "output_value": head_a_uuid})
    if head_b_uuid is not None:
        stack.outputs.append({"output_key": "head_b_uuid", "output_value": head_b_uuid})

    # Set attributes
    for attr, value in six.iteritems(kwargs):
        setattr(stack, attr, value)

    return stack


def decorate_test_methods_with(decorator):
    """Decorator for a TestCase class. Wrap every test method with the `decorator` parameter."""
    def class_decorator(decorated_class):
        for (name, member) in six.iteritems(decorated_class.__dict__):
            if name.startswith("test_") and callable(member):
                setattr(decorated_class, name, decorator(member))
        return decorated_class
    return class_decorator


@contextmanager
def temporary_file(content):
    """Creates a temporary file with a certain content."""
    with NamedTemporaryFile(mode="w+") as temp:
        temp.write(content)
        temp.flush()
        yield temp



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

import tenacity

from clusterondemand.bcm_version import BcmVersion
from clusterondemand.configuration_validation import is_ssh_key_specified
from clusterondemand.exceptions import CODException
from clusterondemand.ssh import SSHAuthenticationError, SSHConnectionError, SSHExecutor, validate_ssh_access
from clusterondemandconfig import ConfigNamespace
from clusterondemandconfig.exceptions import ConfigLoadError

from .cmclient import CMDaemonJSONClient
from .utils import wait_for_socket

log = logging.getLogger("cluster-on-demand")

clusterwaiters_ns = ConfigNamespace("cluster.waiters")
clusterwaiters_ns.add_parameter(
    "wait_ssh",
    default=300,
    help_varname="SECONDS",
    help="Wait up to that many seconds for SSH to come up"
)
clusterwaiters_ns.add_parameter(
    "wait_cmdaemon",
    default=600,
    help_varname="SECONDS",
    help="Wait up to that many seconds for CMDaemon to come up"
)
clusterwaiters_ns.add_switch_parameter(
    "start_nodes",
    default=False,
    help="Power on compute nodes after starting the cluster"
)
clusterwaiters_ns.add_parameter(
    "wait_for_nodes",
    default=0,
    help_varname="SECONDS",
    help="Wait for up to that many seconds for the compute nodes to come up"
)
clusterwaiters_ns.add_validation(lambda param, config: validate_waiters(config))


def wait_for_ssh(floating_ip, timeout):
    log.info("Waiting for sshd to start (ssh root@%s)." % floating_ip)
    wait_for_socket(floating_ip, 22, timeout, throw=True)

    log.debug("Waiting for {secs:.0f} seconds.".format(secs=timeout))
    try:
        with SSHExecutor(floating_ip, connect_timeout=timeout):
            pass
    except SSHConnectionError:
        raise CODException(f"Timed out while waiting for sshd to start (ssh root@{floating_ip}).")
    except SSHAuthenticationError:
        # SSHd started successfully. Caller should use validate_ssh_access function to check access
        pass
    log.debug("sshd initialized. Finished waiting.")
    return


def wait_for_cmdaemon_to_start(floating_ip, timeout, cmdaemon_version=None, only_sock=False):
    log.info("Waiting for CMDaemon to start.")
    wait_for_socket(floating_ip, 8081, timeout, throw=True)
    if only_sock:
        return None

    client = CMDaemonJSONClient(floating_ip, cmdaemon_version=cmdaemon_version)
    ready_call_timeout = 10

    timed_out_msg = "Timed out while waiting for CMDaemon to initialize."

    @tenacity.retry(
        stop=tenacity.stop_after_delay(timeout),
        retry=tenacity.retry_if_exception_message(timed_out_msg),
        wait=tenacity.wait_fixed(1),
        before_sleep=tenacity.before_sleep_log(log, logging.DEBUG),
        reraise=True,
    )
    def client_wait():
        services = ("dhcpd", "named", "nfs", "shorewall", "RamdiskService",
                    "SysConfigGenerator", "SysConfigWriter")
        if not client.ready(ready_call_timeout, services=services):
            raise CODException(timed_out_msg)

    log.info("Waiting for CMDaemon to initialize.")
    client_wait()
    log.debug("CMDaemon initialized. Finished waiting.")
    return client


def wait_for_compute_nodes(cmclient, timeout, node_names=None):
    try:
        nodes = {n["uniqueKey"]: n for n in cmclient.getComputeNodes()
                 if node_names is None or n["hostname"] in node_names}
    except Exception as e:
        raise Exception("Failed to get compute nodes" +
                        (f" {node_names} " if node_names else " ") +
                        "to wait for") from e

    if not nodes:
        log.debug("No compute nodes to wait for")
        return

    log.info(f"Waiting for compute nodes {[n['hostname'] for n in nodes.values()]} to start.")

    if BcmVersion(cmclient.call("cmmain", "getVersion")["cmVersion"]) > "8.2":
        log.debug("Using new device status API to wait for nodes.")
        api_service = "cmstatus"
        api_method = "getStatusForDevices"
    else:
        log.debug("Using old device status API to wait for nodes.")
        api_service = "cmdevice"
        api_method = "getStatusForDeviceArray"

    node_keys = [int(k) for k in nodes.keys()]
    remaining_nodes = nodes
    end_time = time.time() + timeout
    while time.time() < end_time:
        log.debug("Fetching node status from CMDaemon.")
        statuses = cmclient.call(api_service, api_method, [node_keys])

        remaining_nodes = []
        stopped_nodes = []
        for status in statuses:
            node = nodes[status["refDeviceUniqueKey"]]
            status_name = status["status"]
            log.debug(f"Node {node['hostname']} is {status_name}")
            if status_name != "UP":
                remaining_nodes.append(node)
            if status_name == "DOWN":
                stopped_nodes.append(node)

        if stopped_nodes:
            try:
                log.debug("Some compute nodes became DOWN during powering on, let's try to power"
                          " them on again...")
                cmclient.powerOnNodes(stopped_nodes)
            except Exception:
                log.debug("Failed while retrying power on, probably because previous power on "
                          " operation is still running", exc_info=True)

        if not remaining_nodes:
            break

        remaining_time = end_time - time.time()
        if remaining_time > 0:
            log.debug("Waiting for %d of %d nodes to start, %d:%02d min remaining...",
                      len(remaining_nodes), len(nodes), remaining_time // 60, remaining_time % 60)
            time.sleep(min(10, remaining_time))
    else:
        names = [n["hostname"] for n in remaining_nodes]
        raise CODException(f"Timed out while waiting for nodes {names} to start")


def wait_for_cluster_to_be_ready(config, floating_ip, bcm_version):
    if config["wait_ssh"] > 0:
        wait_for_ssh(floating_ip, config["wait_ssh"])
    else:
        log.info("Not waiting for SSH")

    if config["wait_cmdaemon"] > 0:
        if config["wait_ssh"] > 0:
            validate_ssh_access(floating_ip)
            cmclient = wait_for_cmdaemon_to_start(floating_ip, config["wait_cmdaemon"], bcm_version)
        else:
            # Without SSH access we are not able to download CMDaemon creds to connect to it
            # and wait until it starts properly. So, let's wait until its socket at least.
            wait_for_cmdaemon_to_start(floating_ip, config["wait_cmdaemon"], only_sock=True)

    if config["start_nodes"]:
        assert cmclient, "Use validate_waiters to validate these params"
        cmclient.powerOnNodes(cmclient.getComputeNodes())

    if config["wait_for_nodes"] > 0:
        assert cmclient and config["start_nodes"], "Use validate_waiters to validate these params"
        wait_for_compute_nodes(cmclient, config["wait_for_nodes"])


def validate_waiters(config):
    # Not all configuration parameters are used for all "cod" commands, but some of those commands still need waiters
    # Therefore, we  validate the values of various keys, only if those keys exist in configuration
    if not config.get("run_cm_bright_setup", True) and (config["wait_cmdaemon"] or config["wait_for_nodes"]):
        log.warning("--run-cm-bright-setup=no overrides --wait-cmdaemon and --wait-for-nodes options")
        config["wait_cmdaemon"] = 0
        config["wait_for_nodes"] = 0
    if config["wait_for_nodes"] > 0:
        if ("ssh_key_pair" in config or "ssh_pub_key_path" in config) and not is_ssh_key_specified(config):
            raise ConfigLoadError("Using '--wait-for-nodes' requires path to ssh key to be specified")
        if "wait_cmdaemon" in config and not config["wait_cmdaemon"] > 0:
            raise ConfigLoadError("Using '--wait-for-nodes' requires '--wait-cmdaemon' to also be set")
        if "start_nodes" in config and not config["start_nodes"]:
            raise ConfigLoadError("Using '--wait-for-nodes' requires '--start-nodes' to also be set")



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import yaml
from itertools import chain
from logging import getLogger

from six.moves.urllib.request import urlopen

from .image import Image

log = getLogger("cluster-on-demand")


class ManifestVersionMismatch(Exception):
    def __init__(self, version):
        super(ManifestVersionMismatch, self).__init__(
            "Version %d is not supported.", version
        )


class Repository(object):
    """
    Image repository object.

    That class represents image repo list of version 1.
    """

    @classmethod
    def read_manifest_file(cls, manifest_url):
        yaml_stream = urlopen(manifest_url)
        try:
            return yaml.safe_load(yaml_stream)
        finally:
            yaml_stream.close()

    @classmethod
    def bootstrap_manifest(cls):
        return {"version": 1}

    @classmethod
    def add_image(cls, repo_data, image_data):
        images = repo_data.get("images", [])
        images.append(image_data)
        repo_data["images"] = images
        return repo_data

    @classmethod
    def make_copy_override(cls, source_repo,
                           images=None,
                           image_descriptions=None,
                           links=None,
                           resolve_link_func=None,
                           baseurl=None):
        return cls(
            images=image_descriptions or
            [img.raw_description() for img in images] if images is not None else
            source_repo._images,

            links=links or source_repo._links,
            resolve_link_func=resolve_link_func or source_repo._resolve_link,
            baseurl=baseurl or source_repo._baseurl
        )

    @classmethod
    def make_from_data(cls, data, baseurl=None):
        if data["version"] != 1:
            raise ManifestVersionMismatch(data["version"])

        return cls(data.get("images", []),
                   data.get("links", []),
                   cls.make_from_url,
                   baseurl)

    @classmethod
    def make_from_url(cls, manifest_url):
        """Create an object using the manifest file located by URL."""
        data = cls.read_manifest_file(manifest_url)

        return cls.make_from_data(data, manifest_url)

    def __init__(self, images, links, resolve_link_func, baseurl):
        """
        Construct an object.

        images - list of images
        links - list of objects that refer other repo sources
        resolve_link_func - function that accepts link as an argument and returns a Repository
        object
        """
        self._images = images
        self._links = links
        self._resolve_link = resolve_link_func
        self._baseurl = baseurl

    def raw_data(self):
        return {"version": 1,
                "images": self._images,
                "links": self._links}

    def repos_iter(self):
        """Recursive iterate over all linked repos, including current."""
        yield self
        for link in self._links:
            repo = self._resolve_link(link)
            for int_repo in repo.repos_iter():
                yield int_repo

    def image_descriptions(self):
        return self._images

    def images_noraise(self):
        return (img for img in
                (Image.make_from_descripton_noraise(descr, self._baseurl) for descr in
                 self.image_descriptions())
                if img is not None)

    def images_rec_noraise(repo):
        return chain.from_iterable(
            (one_repo.images_noraise() for one_repo in repo.repos_iter())
        )

    def images(self):
        return (Image.make_from_descripton(descr, self._baseurl)
                for descr in self.image_descriptions())



#!/usr/bin/env python
# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import re
from copy import deepcopy

from clusterondemand.exceptions import CODException

# Regular expression matching the following format:
#   [SRC_CIDR[:SRC_PORT_OR_PORT_RANGE],]DST_PORT_OR_PORT_RANGE[:PROTOCOL]
IPV4_CIDR = r"(\d{1,3}(\.\d{1,3}){3}(?:/\d{1,2})"
PORT_OR_PORT_RANGE = r"\d{1,5}(-\d{1,5})?)?"
PORT = r"\d{1,5}"
PROTOCOL = "(tcp|udp)"

SOURCE_REGEX = "^(" + IPV4_CIDR + "(:" + PORT_OR_PORT_RANGE + ",)?"
DESTINATION_REGEX = "(" + PORT_OR_PORT_RANGE + "(:" + PROTOCOL + ")?)$"

INBOUND_RULE_REGEX = SOURCE_REGEX + DESTINATION_REGEX

ANY = "*"

log = logging.getLogger("cluster-on-demand")


class InboundTrafficRule(object):

    def __init__(self, inbound_rule):

        if not self.validate_inbound_rule(inbound_rule):
            raise CODException("Inbound rule '{inbound_rule}' does not match the expected format.\n"
                               "The expected format is: [src_cidr[:src_port],]dst_port[:protocol]\n"
                               "Examples: '80' '21:udp' '11.0.0.0/24,20-23:TCP' "
                               "'12.0.0.0/32:6000-6500,443'".format(inbound_rule=inbound_rule))
        self.inbound_rule = inbound_rule
        self._parse_rule()

    def _parse_rule(self):
        if "," in self.inbound_rule:
            source_rule, destination_rule = self.inbound_rule.split(",")
        else:
            source_rule = "{src_cidr}:{src_port}".format(src_cidr="0.0.0.0/0", src_port=ANY)
            destination_rule = self.inbound_rule
        self.src_cidr, self.src_first_port, self.src_last_port = self.parse_source(source_rule)
        self.dst_first_port, self.dst_last_port, self.protocol = self.parse_destination(destination_rule)
        if not self.dst_first_port <= self.dst_last_port and self.src_first_port <= self.src_last_port:
            raise CODException("Invalid inbound traffic rule: port_range_min must be <= port_range_max")

    @staticmethod
    def parse_port(port):
        if "-" in port:
            first_port, last_port = port.split("-")
        else:
            first_port = last_port = port
        return first_port, last_port

    def parse_source(self, source_rule):
        if ":" in source_rule:
            cidr, port = source_rule.split(":")
        else:
            cidr, port = source_rule, ANY
        self.src_port = port
        first_port, last_port = self.parse_port(port)
        return cidr, first_port, last_port

    def parse_destination(self, destination_rule):
        if ":" in destination_rule:
            port, protocol = destination_rule.split(":")
        else:
            port, protocol = destination_rule, ANY
        self.dst_port = port
        first_port, last_port = self.parse_port(port)
        return first_port, last_port, protocol

    @staticmethod
    def validate_inbound_rule(inbound_rule):
        return re.match(INBOUND_RULE_REGEX, inbound_rule, re.IGNORECASE)

    def __str__(self):
        return self.inbound_rule

    @staticmethod
    def process_inbound_rules(inbound_rules):
        if not inbound_rules:
            return []
        processed_inbound_rules = deepcopy(inbound_rules)
        for inbound_rule in processed_inbound_rules:
            if any(port != "*" for port in [inbound_rule.src_first_port, inbound_rule.src_last_port]):
                log.warning("Source port was specified in rule {traffic_rule} but is only supported by Azure "
                            "so will be ignored.".format(traffic_rule=inbound_rule))

            # if the protocol is not specified in the rule, it has to be split into two separate rules (tcp/udp)
            if inbound_rule.protocol == "*":
                inbound_rule.protocol = "tcp"
                inbound_rule_copy = deepcopy(inbound_rule)
                inbound_rule_copy.protocol = "udp"
                processed_inbound_rules.append(inbound_rule_copy)

        return processed_inbound_rules



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import os
import secrets
import string
import textwrap
from datetime import datetime

import six
from six.moves import range

import clusterondemand.tracing as tracing
from clusterondemand import const
from clusterondemand.clusternameprefix import ensure_cod_prefix
from clusterondemand.exceptions import CODException
from clusterondemand.tags import format_cluster_tags, format_packagegroups_tags
from clusterondemandconfig import config

log = logging.getLogger("cluster-on-demand")


def generate_random_cluster_password(length):
    assert length > 0
    assert length < 256

    letters = string.ascii_letters if six.PY3 else string.letters
    chars = letters + string.digits
    password = "".join((secrets.choice(chars)) for x in range(length))
    return password


def enable_cmd_debug_commands(subsystems="*", base_paths=None):
    """Return list with the proper commmands to enable debug on logging.cmd.conf"""
    LOG_CONF = "cm/local/apps/cmd/etc/logging.cmd.conf"

    if not base_paths:
        base_paths = ["/", "/cm/images/default-image/"]

    # We want to replace "debug: " for "debug: subsystems", but not on the commented lines
    # that's why "/\\#/"
    return [
        "sed -i '/\\#/! s/debug:.*/debug: {subsystems}/' {path}".format(
            subsystems=subsystems,
            path=os.path.join(base_path, LOG_CONF),
        )
        for base_path in base_paths
    ]


def set_cmd_advanced_config_commands(parameters, base_paths=None):
    """
    Return list with the proper commands to configure specified AdvancedConfig parameters on cmd.conf

    :param parameters: a list of parameters in the ["param1=value1", "param2=value2", ...] format
    :param base_paths: a list of filesystem prefixes
    """
    CMD_CONF = "cm/local/apps/cmd/etc/cmd.conf"

    assert parameters and all(isinstance(p, str) for p in parameters)

    if not base_paths:
        base_paths = ["/", "/cm/images/default-image/"]

    return [
        "! grep -P '^ *AdvancedConfig *=' \"{path}\""  # Make sure it wasn't set before
        " && echo 'AdvancedConfig = {{ {params} }}' >> \"{path}\"".format(
            params=", ".join(f'"{p}"' for p in parameters),
            path=os.path.join(base_path, CMD_CONF),
        )
        for base_path in base_paths
    ]


def cluster_name_generator(head_node_image):
    version = head_node_image.version
    if version == "trunk":
        version = version.replace("trunk", "t")
    else:
        version = "b" + version.replace(".", "")

    distro = head_node_image.distro
    distro = distro.lower().replace("centos", "c")
    distro = distro.lower().replace("sles", "s")
    distro = distro.lower().replace("ubuntu", "u")
    distro = distro.lower().replace("rocky", "r")
    distro = distro.lower().replace("alma", "a")

    now = datetime.now()

    base_name = ensure_cod_prefix(f"{version}-{distro}-{now:%m-%d}")
    yield base_name

    index = 0
    while True:
        index = index + 1
        yield f"{base_name}-{index}"


def make_cluster_name(head_node_image, cluster_exists_func):
    """
    Creates a unique non-existent depending on user input
    """
    if config["name"] is None:
        for name in cluster_name_generator(head_node_image):
            if not cluster_exists_func(name):
                return name
    else:
        name = ensure_cod_prefix(config["name"])
        if cluster_exists_func(name):
            raise CODException(textwrap.dedent(
                "A cluster with the name '%s' already exists. Remember that "
                "you can skip specifying --name, which will result in an autogenerated "
                "unique name for your cluster. " % name
            ))
        return name

    # Unreachable. Gotta make the linter happy
    return None


def generate_uuid_tag(head_node_image):
    return format_cluster_tags({const.COD_HEAD_IMAGE_UUID: head_node_image.uuid})[0]


def generate_cluster_tags(head_node_image, only_image_tags=False):
    tags = [const.COD_TAG]
    tags += format_cluster_tags({
        const.COD_TOOL_VERSION_TAG: const.COD_TOOL_VERSION,
        const.COD_DISTRO_TAG: head_node_image.distro,
        const.COD_VERSION_TAG: head_node_image.version,
        const.COD_CMDREV_TAG: head_node_image.cmd_revision,
        const.COD_HEAD_IMAGE_CREATED_AT: head_node_image.created_at,
        const.COD_HEAD_IMAGE_ID: head_node_image.id,
        const.COD_HEAD_IMAGE_NAME: head_node_image.name,
        const.COD_HEAD_IMAGE_REV: head_node_image.revision,
    })
    tags.append(generate_uuid_tag(head_node_image))

    # On vmware we cannot create any new tags on cluster creation
    # So we leave this separate
    if not only_image_tags:
        tags += format_cluster_tags({
            const.COD_TRACE_ID: tracing.get_trace_id(),
        })

    tags += format_packagegroups_tags(head_node_image.package_groups)
    return tags


def detect_duplicate_tags(first, second):
    for tag1 in first:
        if "=" in tag1:
            match = tag1.split("=")[0]
            for tag2 in second:
                if tag2.startswith(match):
                    return match
        else:
            if tag1 in second:
                return tag1
    return None



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import yaml
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from itertools import filterfalse
from logging import getLogger

import pytz

from clusterondemand.exceptions import CODException
from clusterondemand.imagetable import CODImageAdapter, make_images_table
from clusterondemand.paramvalidation import (
    validate_headnode_image_value,
    validate_image_value,
    validate_node_image_value
)
from clusterondemandconfig import BCM_VERSION, ConfigNamespace
from cmbrightimagerepo.image import Image

from .distro_family import ImageDistroSpec, latest_distro
from .filters.byid import make_id_query, match_by_id_query
from .filters.packagegroups import match_by_package_groups
from .filters.tags import match_by_tags
from .queryparser import classify_ids, only_explicit_ids

log = getLogger("cluster-on-demand")

HEAD_NODE_IMAGE, COMPUTE_NODE_IMAGE = "head node", "node"


@dataclass
class CODImage:
    bcm_optional_info: str = ""
    cloud_type: str = ""
    cmd_revision: int = -1
    created_at: datetime = None
    distro: str = ""
    distro_family: str = ""
    id: str = ""
    name: str = ""
    image_visibility: str = ""
    revision: int = -1
    size: int = 0
    tags: list = ""
    type: str = ""
    uploaded_at: datetime = None
    uuid: str = ""
    version: str = ""
    package_groups: list = field(default_factory=list)
    pkg_hashes: list = field(default_factory=list)
    bcm_api_hash: str = ""
    node_images: list = field(default_factory=list)

    @staticmethod
    def image_name(image_type, image_id, image_revision):
        if image_type == "headnode":
            prefix = "bcmh"
        elif image_type == "node":
            prefix = "bcmn"
        elif image_type == "node-installer":
            prefix = "bcni"
        elif image_type == "edge-iso":
            prefix = "bcm-ni-edge"
        else:
            raise Exception("Unsupported image type %s" % (image_type))
        return f"{prefix}-{image_id}-{image_revision}"

    # TODO Move the code from cmbrightimagerepo.image.Image to here?
    @classmethod
    def from_manifest_data(cls, manifest, baseurl=None):
        image = Image.make_from_descripton(manifest, baseurl)

        # Remove everything after first digit
        # Same logic as in build-cod-image.sh
        distro_family = re.sub(r"\d.*", "", image.os())

        return CODImage(
            bcm_optional_info="",
            cloud_type=image.cloud_type(),
            cmd_revision=image.cmd_revision(),
            created_at=datetime.fromtimestamp(image.created_at(), pytz.utc),
            distro=image.os(),
            distro_family=distro_family,
            id=image.id(),
            name=CODImage.image_name(image.image_type(), image.id(), image.revision()),
            image_visibility="public",
            revision=image.revision(),
            size=image.uncompressed_size(),
            tags=image.tags(),
            type=image.image_type(),
            uploaded_at=datetime.fromtimestamp(image.created_at(), pytz.utc),
            uuid="N/A",
            version=str(image.bcm_version()),
            package_groups=sorted(image.package_groups()),
            bcm_api_hash=image.bcm_api_hash(),
        )

    @classmethod
    def from_manifest_file(cls, filename, baseurl=None):
        try:
            with open(filename) as manifest_source:
                return cls.from_manifest_data(yaml.safe_load(manifest_source), baseurl)
        except Exception as e:
            raise CODException(
                f"Error loading image description from {filename}: {e}",
                caused_by=e
            )

    def get_properties(self):
        """
        Return dictionary with the image properties. Useful for when creating and tagging an image
        """
        return {
            "bcm_cloud_type": self.cloud_type,
            "bcm_cmd_revision": str(self.cmd_revision),
            "bcm_created_at": self.created_at.isoformat(),
            "bcm_distro": self.distro,
            "bcm_distro_family": self.distro_family,
            "bcm_image_id": self.id,
            "bcm_image_revision": str(self.revision),
            "bcm_image_type": self.type,
            "bcm_optional_info": self.bcm_optional_info,
            "bcm_package_groups": ",".join(self.package_groups),
            "bcm_pkg_hashes": ",".join(f"{pkg}={git_hash}" for pkg, git_hash in self.pkg_hashes),
            "bcm_version": str(self.version),
            "bcm_api_hash": str(self.bcm_api_hash),
        }

    def get_tags(self):
        return self.tags


class ImageSource:
    def __init__(self,
                 ids=None,
                 package_groups=None,
                 tags=None,
                 version=None,
                 distro=None,
                 revision=None,
                 status=None,
                 type=None,
                 advanced=False,
                 image_visibility="any",
                 cloud_type="any"):
        assert image_visibility in ["public", "private", "any"]
        # Not sure why this check
        # assert(not type or advanced)

        ids = ["*"] if ids is None else ids

        (self.ids, self.patterns, self.regexes, self.uuids) = classify_ids(ids)
        self.package_groups = package_groups if package_groups is not None else ["any"]
        if self.package_groups == ["none"]:
            self.package_groups = []
        self.tags = tags if tags is not None else []
        self.version = version
        self.distro_spec = ImageDistroSpec(distro)
        self.distro = self.distro_spec.full
        self.distro_family = None if self.distro else self.distro_spec.family
        self.revision = revision
        self.status = status
        self.type = type
        self.advanced = advanced
        self.image_visibility = image_visibility
        self.cloud_type = cloud_type if cloud_type != "any" else None

    @classmethod
    def from_config(cls, config, ids=None):
        """Instantiate a proper ImageSource from a cod config
        Pick images doesn't have config["ids"]. It's passed by parameter
        So the implemenation has to check if the parameter is None.
        Small limitation of this interface
        """
        raise NotImplementedError()

    def _iter_from_source(self):
        """Return generator with the specified images"""
        raise NotImplementedError()

    @classmethod
    def find_images_using_options(cls, config):
        source = cls.from_config(config)
        return source.find_images(config=config)

    def find_images(self, latest=None, latest_invert=False, config=None):
        if config is not None:
            if config["all_revisions"]:
                latest = None
                latest_invert = False
            elif config["except_latest"] is not None:
                latest = config["except_latest"]
                latest_invert = True
            else:
                latest = config["latest"]
                latest_invert = False

        assert self.advanced or not self.uuids

        images = self._iter_from_source()

        def log_skip_image(image, fmt, *args, **kwargs):
            log.debug(("{0.id}:{0.revision} is skipped. Cause: " + fmt)
                      .format(image, *args, **kwargs))
            return False
        image_types = ("headnode", "node", "node-installer", "edge-iso")
        images = (
            image for image in images
            if image.type in image_types or
            log_skip_image(
                image, "image type mismatch: {0.type} is not of the possible image types: '{1}'",
                "', '".join(image_types)
            )
        )

        if not self.advanced:
            def id_key_fun(image):
                return image.id
        else:
            def id_key_fun(image):
                return image.name

        only_by_id = only_explicit_ids(self.ids, self.patterns, self.regexes, self.uuids, self.advanced)

        rev_regexes = make_id_query(self.ids, self.patterns, self.regexes)

        images = (
            image for image in images
            if match_by_id_query(
                rev_regexes,
                id_key_fun(image),
                image.revision
            ) or (self.uuids and image.uuid in self.uuids) or
            log_skip_image(image,
                           "image id mismatch: {0.id} do not match either these regexes: {1}, "
                           "or these uuids: {2}",
                           [(revision, regex.pattern) for revision, regex in rev_regexes],
                           self.uuids)
        )

        def valid_distro(image):
            return (
                self.distro_spec is None or image.distro in self.distro_spec or
                log_skip_image(image, "distro mismatch: {0.distro} not in spec {1}", self.distro_spec)
            )

        if not only_by_id:
            images = (
                image for image in images
                if self.revision is None or image.revision == self.revision or
                log_skip_image(image, "revision mismatch: {0.revision} != {1}", self.revision)
            )
            images = (
                image for image in images
                if self.package_groups is None or match_by_package_groups(self.package_groups, image.package_groups) or
                log_skip_image(image, "package groups mismatch: {0.package_groups} != {1}",
                               self.package_groups)
            )
            images = (
                image for image in images
                if valid_distro(image)
            )
            images = (
                image for image in images
                if match_by_tags(self.tags, image.tags) or
                log_skip_image(image, "tags mismatch: {0.tags} do not match requested: {1}", self.tags)
            )
            images = (
                image for image in images
                if self.version is None or image.version == self.version or
                log_skip_image(image, "version mismatch: {0.version} != {1}", self.version)
            )
            images = (
                image for image in images
                if self.image_visibility == "any" or image.image_visibility == self.image_visibility or
                log_skip_image(image, "public visibility mismatch: {0.image_visibility} != {1}", self.image_visibility)
            )

        # cloud_type filtering is out of the 'if' above because we always to do it, even if the user types a specific id
        images = (
            image for image in images
            if self.cloud_type is None or image.cloud_type == self.cloud_type or
            log_skip_image(image, "cloud type mismatch: {0.cloud_type} != {1}", self.cloud_type)
        )

        if latest is not None:
            images = list(images)
            if self.advanced:
                latest_revisions = get_latest_revisions(latest, latest_invert, images)
            else:
                latest_revisions = get_latest_revisions(latest, latest_invert,
                                                        [i for i in images if i.type == "headnode"])
            images = (
                image for image in images
                if image.revision in latest_revisions.get(image.id, [])
            )

        def create_cod_image(head, nodes):
            head.node_images = nodes
            return head

        if not self.advanced:
            images = (create_cod_image(head, nodes) for head, nodes in group_images(images))
        else:
            images = (create_cod_image(image, []) for image in images)

        return images

    @classmethod
    def pick_head_node_image_using_options(cls, opts):
        head_image_id = opts.get("head_node_image")
        image_id = opts.get("image")

        if head_image_id:
            image_ids = [head_image_id, _image_id_from_image_set_id(head_image_id, HEAD_NODE_IMAGE)]
            return cls._pick_image_by_ids(image_ids, HEAD_NODE_IMAGE, opts["cloud_type"])
        elif image_id:
            image_ids = [_image_id_from_image_set_id(image_id, HEAD_NODE_IMAGE)]
            return cls._pick_image_by_ids(image_ids, HEAD_NODE_IMAGE, opts["cloud_type"])
        else:
            # User did not specify head image explicitly, find one based on version, distro, etc.
            return cls._pick_head_node_image_by_opts(opts)

    @classmethod
    def pick_compute_node_image_using_options(cls, opts):
        node_image_id = opts["node_image"]
        image_id = opts["image"]

        if node_image_id == "none":
            return None
        elif node_image_id:
            image_ids = [node_image_id, _image_id_from_image_set_id(node_image_id, COMPUTE_NODE_IMAGE)]
            return cls._pick_image_by_ids(image_ids, COMPUTE_NODE_IMAGE, opts["cloud_type"])
        elif image_id:
            image_ids = [_image_id_from_image_set_id(image_id, COMPUTE_NODE_IMAGE)]
            return cls._pick_image_by_ids(image_ids, COMPUTE_NODE_IMAGE, opts["cloud_type"])
        else:
            # User did not specify node image explicitly, find one based on version, distro, etc.
            return cls._pick_compute_node_image_by_opts(opts)

    @classmethod
    def _pick_head_node_image_by_opts(cls, opts):
        """
        Pick the head node image, based on the configuration (opts)
        Image distribution is decided based on the configuration parameters: distro and distro_default. They can be
        specified from the cli or configuration file, E.g. krusty.ini
        Decision logic:
        default_distro  distro
        Y               Y           use distro
        Y               N           iterate over default_distro, use the first distro for which image is found
        N               Y           use distro
        N               N           get images without filtering them by distribution and use the latest one
        :param opts: config
        :return: CODImage
        """
        if opts["distro_default"] and not opts["distro"]:
            for distro in opts["distro_default"]:
                log.debug(f"Searching for an image for {distro} from distro_default list")
                opts["distro"] = distro
                source = cls.from_config(opts, ids=["*"])
                images = list(source.find_images(latest=True))
                if images:
                    break
        else:
            source = cls.from_config(opts, ids=["*"])
            images = list(source.find_images(latest=True))

        selected_distro = latest_distro([i.distro for i in images], opts["distro"])
        images = [i for i in images if i.distro == selected_distro]

        if 0 == len(images):
            raise NoImagesFoundError.create_for_filter_options(opts, HEAD_NODE_IMAGE)
        elif 1 == len(images):
            return images[0]
        else:
            raise MultipleImagesFoundError.create_for_filter_options(opts, images, HEAD_NODE_IMAGE)

    @classmethod
    def _pick_compute_node_image_by_opts(cls, opts):
        # Select a compute node image based on the --version/--distro/etc
        # This is done by selecting a head node image, and then picking the one was matched to it

        node_head_image = cls._pick_head_node_image_by_opts(opts)

        if not node_head_image or not node_head_image.node_images:
            return None
        else:
            node_images = [img for img in node_head_image.node_images if img.type == "node"]
            assert len(node_images) == 1, "One node image expected, %d found" % len(node_images)
            return node_images[0]

    @classmethod
    def _pick_image_by_ids(cls, image_ids, image_type, cloud_type):
        """Iterate over list of Glance image ids in order. Return the first image found with such an image id.

        When a tried image id maps to multiple images, an error is raised. If multiple image ids in the
        parameter map to an image, the first image is returned. If none of the image ids map to an
        image, an error is raised as well.
        """
        for image_id in image_ids:
            source = cls(ids=[image_id], advanced=True, cloud_type=cloud_type)
            found_images = list(source.find_images())

            if 0 == len(found_images):
                pass
            elif 1 == len(found_images):
                return found_images[0]
            else:
                raise MultipleImagesFoundError.create_for_image_id(image_id, found_images, image_type)
        else:
            raise NoImagesFoundError.create_for_tried_image_ids(image_ids, image_type)


def _image_id_from_image_set_id(image_set_id, image_type):
    """
    Convert an <image set id>:<revision>, into the id of a Glance image

    >>> _image_id_from_image_set_id("centos7u5-trunk:213", HEAD_NODE_IMAGE)
    'bcmh-centos7u5-trunk-213'
    >>> _image_id_from_image_set_id("ubuntu1804-9.0-dev:1", COMPUTE_NODE_IMAGE)
    'bcmn-ubuntu1804-9.0-dev-1'
    """
    prefix = {HEAD_NODE_IMAGE: "bcmh", COMPUTE_NODE_IMAGE: "bcmn"}[image_type]
    return prefix + "-" + re.sub(r":([0-9]+)$", r"-\1", image_set_id)


class NoImagesFoundError(CODException):
    """Raised when no image could be found for either the uuid, image ids or filter."""

    @classmethod
    def create_for_tried_image_ids(cls, tried_ids, image_type):
        tried_ids = list(set(tried_ids))
        if 1 == len(tried_ids):
            return cls("Could not find a %s image for %s" % (image_type, tried_ids[0]))
        else:
            return cls("Could not find a %s image for any of these ids: %s" % (image_type, ", ".join(tried_ids)))

    @classmethod
    def create_for_filter_options(cls, filter_options, image_type):
        return cls("Could not find any %s image matching these criteria:\n%s" %
                   (image_type, _format_filter_string(filter_options)))


class MultipleImagesFoundError(CODException):
    """Raised when the specified image id, or filter options, generated multiple images to chose from."""

    @classmethod
    def create_for_image_id(cls, image_id, images, image_type):
        return cls(
            'More than one %s image matched "%s". Use %s <UUID> to explicitly chose one:\n%s' %
            (image_type, image_id, _flag_for_image_type(image_type), _format_image_list(images))
        )

    @classmethod
    def create_for_filter_options(cls, filter_options, images, image_type):
        return cls(
            "There were several %s images that match these criteria, use %s <UUID> to explicitly chose one:\n%s\n%s\n" %
            (image_type, _flag_for_image_type(image_type), _format_filter_string(filter_options),
             _format_image_list(images))
        )


def _flag_for_image_type(image_type):
    return {HEAD_NODE_IMAGE: "--head-node-image", COMPUTE_NODE_IMAGE: "--node-image"}[image_type]


def _format_filter_string(opts):
    filter_options = ["tags", "version", "distro", "package_groups", "image_visibility"]
    option_and_values = ((opts.get_item_for_key(name).parameter, opts[name]) for name in filter_options)
    filter_string = "\n".join(
        "{namespace}.{config_option}={val}".format(
            namespace=opt.namespaces[-1],
            config_option=opt.key,
            val=val
        )
        for opt, val in option_and_values
    )
    return filter_string


def _format_image_list(images):
    return make_images_table(
        (CODImageAdapter(img) for img in images),
        columns=["id_revision", "name", "uuid"],
        sortby=["id_revision"],
        output_format="table"
    )


def get_latest_revisions(n, invert, images):
    """
    Return a dictonary {<image-id>: [n latest revisions]}

    If `invert`, we return the opposite revisions (all revisions except the last n)
    """
    image_revisions = defaultdict(set)
    for image in images:
        image_revisions[image.id].add(image.revision)

    def last_n(images):
        return sorted(images, reverse=True)[n:] if invert else sorted(images, reverse=True)[:n]

    return {
        image_id: last_n(revisions)
        for image_id, revisions in image_revisions.items()
    }


def group_all_by(source, keyfunc, ismainfunc):
    """Group item list to the [(mainitem, [item,...]), ...] structure, preserving the sort order.

    >>> DATA = [(1, "m", 1), (3, "i", 2), (2, "i", 1), (4, "m", 3)]
    >>> ismainfunc = lambda i: i[1] == "m"
    >>> keyfunc = lambda i: i[2]
    >>> list(group_all_by(DATA, keyfunc, ismainfunc))
    [((1, 'm', 1), [(2, 'i', 1)]), ((4, 'm', 3), [])]

    """
    source = list(source)
    for main_item in filter(ismainfunc, source):
        key = keyfunc(main_item)
        items = filterfalse(ismainfunc, source)
        related_items = filter(
            lambda item: keyfunc(item) == key,
            items
        )
        yield (main_item, list(related_items))


def group_images(images):
    return group_all_by(
        images,
        keyfunc=lambda image: (image.id, int(image.revision)),
        ismainfunc=lambda image: image.type == "headnode"
    )


findimages_ns = ConfigNamespace("common.images.find", help_section="image filter parameters")
findimages_ns.add_parameter(
    "cloud_type",
    help="Filter images by cloud type. Set 'any' for any cloud type. Useful with --advanced"
)
findimages_ns.add_parameter(
    "version",
    help="(default: all versions) show only images from this version"
)
findimages_ns.add_parameter(
    "status",
    default="active",
    help="Status of the image"
)
findimages_ns.add_parameter(
    "distro",
    help=("(default: all distros) show only images from this distro."
          "This is a regex")
)
findimages_ns.add_enumeration_parameter(
    "distro_default",
    default=[],
    help="List of distributions. If the user didn't explicitly specify a distro, "
         "iterate over this list and use the first distribution, for which the image is found. "
         "Ignored if combined with --distro. "
         "Each list element is a regex"
)
findimages_ns.add_parameter(
    "revision",
    type=int,
    help="Select images with this revision."
)
findimages_ns.add_enumeration_parameter(
    "package_groups",
    help_varname="PACKAGE_GROUP",
    default=["any"],
    help=("Select images with these package groups installed "
          "(use 'none' to pick an image with no package groups installed, "
          " use 'any' (default) to pick an image with any set of package groups installed.")
)
findimages_ns.add_enumeration_parameter(
    "tags",
    default=[],
    help=("List of tags, the images must have them all. (AND). "
          "--tags any will pick images with any set of tags")
)
findimages_ns.add_repeating_positional_parameter(
    "ids",
    default=["*"],
    help=("List of image specifiers. By default (unless '--advanced' is used) those are "
          "assumed to be in the COD 'ImageID[:revision]' format. Each specifier can be a "
          "plain string (e.g. centos7u5-8.0), a wildcard (e.g. centos*), or a regex (e.g. "
          "centos7u[45]*). It is also possible to explicitly suffix any of the above with "
          "a colon, followed by a number specifying the image revision (e.g. 'centos7u5-8.0:42', "
          "or 'centos*:42', or '*:42'). If the '--advanced' flag is used, the specifiers "
          "will not be treated as COD ImageIDs, but instead they will be treated as "
          "'Name' or 'UUID' strings (or wildcards, or regexes). For example 'bcmh-centos*'.")
)
findimages_ns.add_switch_parameter(
    "all_revisions",
    help="Pick all revisions for each image ID"
)
findimages_ns.add_parameter(
    "latest",
    default=1,
    help="Pick this many latest images for each image ID"
)
findimages_ns.add_switch_parameter(
    "advanced",
    help="Filter image by its name or UUID rather than ImageID. "
    "Will also change listing output."
)
findimages_ns.add_parameter(
    "except_latest",
    type=int,
    help="Pick all images besides this number of the latest images for each image ID"
)
findimages_ns.add_parameter(
    "image_visibility",
    choices=["any", "public", "private"],
    default="any",
    advanced=True,
    help="Whether to show 'any' images (both public and private), "
         "or only 'public' or 'private' images. Defaults to 'any'."
)

pickimages_ns = ConfigNamespace("common.images.pick", help_section="image selection parameters")
pickimages_ns.import_namespace(findimages_ns)
pickimages_ns.remove_imported_parameter("ids")
pickimages_ns.override_imported_parameter("package_groups", default=["none"])
pickimages_ns.override_imported_parameter("version", default=BCM_VERSION)
pickimages_ns.add_parameter(
    "image",
    help_varname="IMAGE_SPEC",
    help=("Single image selector statement. "
          "See cm-cod-os image list --help. "
          "Overrides filter arguments such as --version, --distro, etc."),
    validation=validate_image_value
)
pickimages_ns.add_parameter(
    "head_node_image",
    help_varname="UUID|IMAGE-NAME|IMAGE-SET",
    help=("Single image selector statement for the head node image. Can either be an image "
          "UUID, the name of that image or the name of the image set. Overrides the head node "
          "image selected by --image and all other image filter arguments."),
    validation=validate_headnode_image_value
)
pickimages_ns.add_parameter(
    "node_image",
    help=("Single image selector statement for node image (as in advanced mode) "
          "'--node-image none' will force the cluster to not use a node image at all. "
          "Overrides the head node image selected by --image and all other image "
          "filter arguments."),
    validation=validate_node_image_value
)



#!/usr/bin/env python
# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

from setuptools import find_packages, setup

try:
    from clusterondemand._version import version as __version__
except ImportError:
    from setuptools_scm import get_version
    __version__ = get_version(
        root="..",
        relative_to="__file__",
        write_to="cluster-on-demand/clusterondemand/_version.py"
    )

with open("README.md", "r") as file_in:
    readme = file_in.read()

setup(
    name="cm-cluster-on-demand",
    version=__version__,
    description="Bright Cluster on Demand Utility",
    author="Cloudteam",
    author_email="cloudteam@brightcomputing.com",
    url="https://www.brightcomputing.com/",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Topic :: System :: Clustering",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Systems Administration"
    ],
    packages=find_packages(),
    package_data={"clusterondemand": ["eula.txt", "version-info.yaml"]},
    tests_require=["pytest", "pycoverage"],
    entry_points={
        "console_scripts": [
            "cm-cod = clusterondemand.cli:cli_main",
        ]
    },
    python_requires=">=3.9",
    install_requires=[
        "setuptools_scm>=4.1.2",
        "paramiko>=2.11.0",
        "netaddr>=0.8.0",
        "pytz>=2022.5",
        "python_dateutil>=1.5",
        "pyyaml>=6.0",
        "PrettyTable>=3.4.0",
        "requests",
        "filelock>=2.0.8",
        "tenacity>=6.2.0",
        "cm-cluster-on-demand-config==" + __version__
    ],
    setup_requires=["setuptools_scm>=4.1.2"]
)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

import os
import yaml
from collections import OrderedDict

import six

PATH_TO_VERSION_INFO = os.path.join(os.path.dirname(os.path.realpath(__file__)), "version-info.yaml")

HUMAN_READABLE_INFO = OrderedDict([
    ("version", "Version"),
    ("branch", "Branch"),
    ("git_hash", "Git Hash"),
    ("git_date", "Git Date"),
    ("build_date", "Build Date"),
])


def get_version_info():
    """Returns dictionary containing the version information."""
    try:
        return yaml.safe_load(open(PATH_TO_VERSION_INFO).read())
    except FileNotFoundError:
        return {
            "version": "trunk",
            "branch": "",
            "git_hash": "",
            "git_date": "",
            "build_date": ""
        }


def print_version_info():
    version_info = get_version_info()
    for key, human_text in six.iteritems(HUMAN_READABLE_INFO):
        value = version_info.get(key)
        if value:
            print("{}: {}".format(human_text, value))



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import re
import subprocess

from .exceptions import UserReportableException

SSH_KEYGEN_REGEX = r"(?P<length>[0-9]+) [^:]+:(?P<fingerprint>\S+) .* \((?P<type>[^\)]+)\)$"


class CODSSHPrivateKeys(object):

    def __init__(self):

        self.keys = {}

        default_keyfiles = [
            "id_dsa",
            "id_ecdsa",
            "id_ed25519",
            "id_rsa"
        ]

        fingerprint_hash = "md5"

        for keyfile_basename in default_keyfiles:
            keyfile = os.path.expanduser("~/.ssh/{}".format(keyfile_basename))
            self._add_key_to_dict(keyfile, fingerprint_hash)

        agent_keys = self._get_ssh_agent_keys(fingerprint_hash)
        if agent_keys:
            self.keys.update(agent_keys)

    def fingerprint_exists(self, check_fingerprint):
        return check_fingerprint in self.keys

    def _add_key_to_dict(self, keyfile, fingerprint_hash):
        if os.path.exists(keyfile):
            ssh_keygen_output = _run_ssh_keygen(keyfile, fingerprint_hash)

            if ssh_keygen_output:
                key_length, key_fingerprint, key_type = ssh_keygen_output
                self.keys[key_fingerprint] = {
                    "type": key_type,
                    "length": int(key_length)
                }

    @staticmethod
    def _get_ssh_agent_keys(fingerprint_hash):
        if "SSH_AUTH_SOCK" not in os.environ:
            return None

        ssh_agent_keys = {}

        ssh_add_proc = subprocess.Popen(["ssh-add", "-E", fingerprint_hash, "-l"],
                                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = ssh_add_proc.communicate()
        if stdout:
            lines = stdout.decode("utf-8").split("\n")
            for line in lines:
                ssh_add_match = re.match(SSH_KEYGEN_REGEX, line)
                if ssh_add_match:
                    key_length = ssh_add_match.group("length")
                    key_fingerprint = ssh_add_match.group("fingerprint")
                    key_type = ssh_add_match.group("type")
                    ssh_agent_keys[key_fingerprint] = {
                        "type": key_type,
                        "length": int(key_length)
                    }

        return ssh_agent_keys


class CODSSHPubKey(object):

    def __init__(self, keyfile, fingerprint="sha256"):

        self.ssh_keygen_success = False
        self.key_length = None
        self.key_fingerprint = None
        self.key_type = None

        ssh_keygen_output = _run_ssh_keygen(keyfile, fingerprint)

        if ssh_keygen_output:
            self.ssh_keygen_success = True
            self.key_length, self.key_fingerprint, self.key_type = ssh_keygen_output

    def is_valid(self):
        return self.ssh_keygen_success


def _run_ssh_keygen(keyfile, fingerprint):
    ssh_keygen_proc = subprocess.Popen(["ssh-keygen", "-E", fingerprint, "-l", "-f", keyfile],
                                       stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = ssh_keygen_proc.communicate()
    if stdout:
        ssh_keygen_match = re.match(SSH_KEYGEN_REGEX, stdout.decode("utf-8"))
        if ssh_keygen_match:
            key_length = ssh_keygen_match.group("length")
            key_fingerprint = ssh_keygen_match.group("fingerprint")
            key_type = ssh_keygen_match.group("type")
            return int(key_length), key_fingerprint, key_type

    return None


def validate_ssh_pub_key(parameter, configuration, allowed_types=None):
    keyfile = configuration[parameter.key]

    if keyfile:

        if not os.path.exists(keyfile):
            raise UserReportableException("SSH key file {} does not exist".format(keyfile))

        pubkey = CODSSHPubKey(keyfile)

        if not pubkey.is_valid():
            raise UserReportableException("SSH key file {} is not a valid key file".format(keyfile))

        if allowed_types and (pubkey.key_type not in allowed_types
                              or pubkey.key_length < allowed_types[pubkey.key_type]):
            allowed_types_list = list(map(lambda key: "{}:{}".format(key, allowed_types[key] or "any"), allowed_types))
            raise UserReportableException("SSH key file {} is of type {} and bitsize {}, "
                                          "permitted types and minimum bitsize are {}".format(
                                              keyfile,
                                              pubkey.key_type,
                                              pubkey.key_length,
                                              ", ".join(allowed_types_list)
                                          ))



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import traceback


class CODException(Exception):
    """Exceptions raised inside COD with some useful user readable message

    :param caused_by: If the exception was raised inside some except block
        the original exception can be specified here for better logging later
    """
    def __init__(self, message, caused_by=None):
        super(CODException, self).__init__(message)
        self.caused_by = caused_by
        self.caused_by_trace = traceback.format_exc()


class UserReportableException(CODException):
    """Exception class reserved for customer portal error messages."""

    pass


class ValidationException(UserReportableException):
    """Exception class reserved for errors validating parameters.

    All validations errors should be shown to the user, so this type extends UserReportableException
    """

    pass



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import os
from contextlib import contextmanager

from filelock import FileLock

from clusterondemand.utils import is_writable_directory

log = logging.getLogger("cluster-on-demand")


@contextmanager
def SmartLock(lock_file):
    """
    An enhanced FileLock that gracefully disables synchronization
    if the parent directory of the specified lock file does not
    exist or is not writable.
    """
    if is_writable_directory(os.path.dirname(lock_file)):
        with FileLock(lock_file):
            yield

    else:
        log.debug("%s is not a writable directory! Locking not available.", os.path.dirname(lock_file))
        yield



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, annotations, print_function

import logging
import pdb
import string
import sys
from itertools import chain
from typing import Any, Callable, Optional

from clusterondemand import tracing
from clusterondemand.exceptions import CODException, UserReportableException
from clusterondemandconfig import (
    UnknownCLIArgumentError,
    UnknownConfigFileParameterError,
    check_for_unknown_config_parameters,
    determine_invoked_command,
    explain_parameter,
    global_configuration,
    human_readable_config_dump,
    load_boot_configuration_for_command,
    load_configuration_for_command,
    print_help,
    print_help_for_positionals_missing_required_value,
    print_help_for_unknown_parameters,
    validate_configuration
)
from clusterondemandconfig.command_context import Command
from clusterondemandconfig.configuration import CommandConfiguration
from clusterondemandconfig.exceptions import ConfigLoadError

from . import accept_eula
from .configuration import get_enforcing_config_files, get_system_config_files
from .import_utils import load_module_attribute
from .setup_logging import setup_logging, temporarily_logging_all_statements_to_syslog
from .utils import setup_signal_handlers
from .version import print_version_info

log = logging.getLogger("cluster-on-demand")


def run_invoked_command(
        command_context,
        pre_run_check: Optional[Callable[[Command, CommandConfiguration], Any]] = None,
        additional_included_config_files=None,
        additional_system_config_files=None):
    """Pass the control to the appropriate tool, specified in 'name' argument.

    Handles the exception, and stack trace printing
    """
    setup_logging()

    with temporarily_logging_all_statements_to_syslog(log):
        log.debug(" ".join(sys.argv))

    non_ascii_chars_used = _check_for_non_ascii_chars_in_arguments()

    # FIXME: clusterondemandconfig doesn't allow us to have this in a clean way
    if "--version" in sys.argv and len(sys.argv) == 2:
        print_version_info()
        sys.exit(0)

    system_config_files = get_system_config_files(additional_included_config_files, additional_system_config_files)
    enforcing_config_files = get_enforcing_config_files()
    (command, clean_sys_argv) = determine_invoked_command(command_context)
    try:
        with sys_argv(clean_sys_argv):
            boot_configuration = load_boot_configuration_for_command(command)

            with global_configuration(boot_configuration):
                configuration = load_configuration_for_command(
                    command, system_config_files=system_config_files, enforcing_config_files=enforcing_config_files
                )
                if configuration["check_config"]:
                    _check_for_unknown_config_parameters(configuration,
                                                         system_config_files=system_config_files,
                                                         enforcing_config_files=enforcing_config_files,
                                                         parameters=command.parameters)

    except UnknownCLIArgumentError as e:
        print_help_for_unknown_parameters(command, e.flags)
        sys.exit(1)
    except CODException as e:
        _log_cod_exception(e, configuration)
        sys.exit(1)

    try:
        if non_ascii_chars_used and configuration["only_ascii_arguments"]:
            raise UserReportableException("Aborting because of a non-ASCII character in an argument. You can use "
                                          "'--no-only-ascii-arguments' to ignore this error.")

        if configuration["help"] or configuration["advanced_help"]:
            print_help(command, configuration)
        elif configuration["explain"]:
            explain_parameter(command, command_context, configuration["explain"])
        elif configuration["show_configuration"]:
            print(human_readable_config_dump(configuration, show_secrets=configuration["show_secrets"]))
        elif configuration.required_positionals_with_missing_values():
            print_help_for_positionals_missing_required_value(command, configuration)
            sys.exit(1)
        elif command.require_eula and not accept_eula.user_accepts_eula(configuration):
            sys.exit(1)
        else:
            try:
                validate_configuration(configuration)
                configuration.lock()
            except RuntimeError as e:
                print(e)
                sys.exit(1)
            if pre_run_check:
                pre_run_check(command, configuration)
            setup_signal_handlers()
            with global_configuration(configuration):
                command_name = command.group.name + " " + command.name
                with tracing.trace_events(command_name):
                    command.run_command()
    except UserReportableException as e:
        _post_mortem(configuration)
        _log_user_reportable_exception(e, configuration)

        sys.exit(1)
    except (ConfigLoadError, CODException) as e:
        _post_mortem(configuration)
        _log_cod_exception(e, configuration)
        # TODO If it's useful, CODException could contain the error code
        sys.exit(1)
    except Exception as e:
        _post_mortem(configuration)
        # TODO Notify us that this happened somehow
        log.error("Unhandled exception")
        # TODO (CM-23030) Replace the next lines for `raise e` in future
        if not configuration["verbose"]:
            log.error("(use '-v' to show the stack trace)")
        _print_error_stack_on_failure(e, configuration)
        log.error("%s: %s" % (e.__class__.__name__, e))
        sys.exit(1)


def _check_for_non_ascii_chars_in_arguments():
    found = False

    for arg in sys.argv:
        if not all(char in string.printable for char in arg):
            found = True
            log.warning("Argument '{argument}' contains a non-ASCII character. This might cause "
                        "unforeseen behavior.".format(argument=arg))

    return found


def _check_for_unknown_config_parameters(configuration, system_config_files, enforcing_config_files,
                                         parameters=None):
    def load_module_parameters(module_name, commands_name):
        commands = load_module_attribute(module_name, commands_name)
        return [] if commands is None else commands.parameters()

    check_all_commands = configuration["check_config_for_all_commands"]
    if parameters is None or check_all_commands:
        aws_parameters = load_module_parameters("clusterondemandaws.cli", "aws_commands")
        azure_parameters = load_module_parameters("clusterondemandazure.cli", "azure_commands")
        openstack_parameters = load_module_parameters("clusterondemandopenstack.cli", "openstack_commands")
        vmware_parameters = load_module_parameters("clusterondemandvmware.cli", "vmware_commands")
        parameters = chain(aws_parameters, azure_parameters, openstack_parameters, vmware_parameters)

    try:
        check_for_unknown_config_parameters(
            parameters,
            system_config_files=system_config_files,
            enforcing_config_files=enforcing_config_files,
            ignore_unknown_namespaces=not check_all_commands,
        )
    except UnknownConfigFileParameterError:
        raise CODException("Some configuration files have problems. Use --no-check-config"
                           " to disable this check.")


def _print_error_stack_on_failure(e, configuration):
    # Print the error stack on failure even if verbosity wasn't requested
    if configuration["print_stack_trace_on_failure"]:
        log.info(e, exc_info=True)
    else:
        log.debug(e, exc_info=True)


def _post_mortem(configuration):
    if configuration["pdb_on_error"]:
        pdb.post_mortem()


def _log_cod_exception(exception, config):
    _print_error_stack_on_failure(exception, config)
    if getattr(exception, "caused_by", None):
        log.debug("Caused by: %s \n %s", exception.caused_by, exception.caused_by_trace)
    log.error(str(exception))


def _log_user_reportable_exception(exception, config):
    _print_error_stack_on_failure(exception, config)
    # We have to show "UserReportableException: " here because that's how the portal identifies the error message
    log.error("UserReportableException: %s" % (exception))


def _determine_thing_to_explain(configuration):
    index = sys.argv.index("--explain")

    if configuration["explain"]:
        return configuration["explain"][0]
    elif index == len(sys.argv) - 1:
        return None
    elif "--" == sys.argv[1 + index]:
        return None
    else:
        return sys.argv[1 + index]


class sys_argv(object):
    """Temporarily replace the contents of sys.argv with the given string."""
    def __init__(self, new_argv):
        self.new_argv = new_argv.strip().split(" ") if isinstance(new_argv, str) else new_argv

    def __enter__(self):
        self.old_argv, sys.argv = sys.argv, self.new_argv

    def __exit__(self, _exception_type, _exception_value, _traceback):
        sys.argv = self.old_argv

    def __call__(self, func):
        def wrapper(_self, *args, **kwargs):
            with self:
                return func(_self, *args, **kwargs)
        return wrapper



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import netaddr

from clusterondemand.cloudconfig import CMD_READY_PROBE_PATH
from clusterondemand.clustercreate import enable_cmd_debug_commands, set_cmd_advanced_config_commands
from clusterondemand.utils import get_commands_to_inject_resolved_hostnames, resolve_hostnames
from clusterondemandconfig import config

log = logging.getLogger("cluster-on-demand")


def add_head_node_commands(cloud_config, internal_ip):
    """ Let's first signal the status wait condition that cloud init started """
    cloud_config.add_status_commands("cloud-init: Started.")

    cloud_config.add_status_commands("cloud-init: Setting up eth0.")
    ip_net = netaddr.IPNetwork(str(config["internal_cidr"]))
    assert internal_ip in ip_net, "%s is not in subnet %s" % (internal_ip, ip_net)

    cloud_config.add_commands([
        "echo Set head node IP",
        "ip link set up dev eth0",
        "ip address flush dev eth0",
        "ip addr add %s/%s dev eth0" % (internal_ip, ip_net.netmask),
    ])

    mtu = config["internal_mtu"]
    if mtu is not None:
        cloud_config.add_status_commands("cloud-init: Setting eth0 custom mtu.")
        cloud_config.add_command("ip link set mtu %d dev eth0" % mtu)

    if config["ssh_password_authentication"]:
        cloud_config.add_status_commands("cloud-init: Enabling password authentication.")
        cloud_config.add_commands([
            "echo 'COD-OS: SSH password authentication requested, editing sshd config file'",
            "sed -i 's/^PasswordAuthentication no/# Bright Cluster on Demand: enabling due to "
            "ssh-password-authentication\\nPasswordAuthentication yes/g' /etc/ssh/sshd_config",
            "echo 'Reloading the sshd service'",
            "if ! systemctl try-reload-or-restart sshd; then",
            "  echo 'Old systemd, using different reload command.'",
            "  systemctl reload-or-try-restart sshd",
            "fi",
        ])

    if config["ssh_pub_key_path"]:
        with open(config["ssh_pub_key_path"]) as pub_key:
            cloud_config.add_command("echo '%s' >> /root/.ssh/authorized_keys" % pub_key.read())

    if config["append_to_root_bashrc"]:
        cloud_config.add_command("echo '# START of entries added by append_to_root_bashrc of Cluster on Demand' "
                                 ">> /root/.bashrc")
        for line in config["append_to_root_bashrc"]:
            cloud_config.add_command("echo '%s' >> /root/.bashrc" % line)
        cloud_config.add_command("echo '# END of entries added by append_to_root_bashrc of Cluster on Demand' "
                                 ">> /root/.bashrc")

    if config["cmd_debug"]:
        subsystems = config["cmd_debug_subsystems"]
        log.debug(f"Setting debug mode on CMDaemon for subsystems: '{subsystems}'")
        for command in enable_cmd_debug_commands(subsystems):
            cloud_config.add_command(command)

    # In this mode the CMD monitoring system will sample metric at randomized intervals
    # And it also says to CMD that hanged compute nodes must be reset
    log.debug("Setting VirtualCluster mode on CMDaemon")
    for command in set_cmd_advanced_config_commands(["VirtualCluster=1"]):
        cloud_config.add_command(command)


def add_run_bright_setup_commands(cloud_config):
    # We have to load these modules so the command cm-bright-setup is available
    # Some older versions don't have the module cm-setup (it's all part of cluster-tools)
    # and would fail to load. That's why "|| true"
    cloud_config.add_commands([
        "echo Loading modules",
        "source /etc/profile.d/modules.sh",
        "module load cluster-tools",
        "module load cm-setup || true",
    ])

    if config["resolve_hostnames"]:
        # This can work around DNS issues by resolving hostnames during cluster creation
        commands = get_commands_to_inject_resolved_hostnames(
            ["/", "/cm/images/default-image"],
            resolve_hostnames(config["resolve_hostnames"]))
        for command in commands:
            cloud_config.add_command(command)

    if config["prebs"]:
        cloud_config.add_status_commands("cloud-init: Running prebs commands.")
        cloud_config.add_command("echo 'Starting custom prebs commands'")
        cloud_config.add_commands(config["prebs"])

    cloud_config.add_status_commands("cloud-init: Running cm-bright-setup.")
    cloud_config.add_command("cm-bright-setup -c /root/cm/cm-bright-setup.conf --on-error-action abort")
    cloud_config.add_status_commands("cloud-init: cm-bright-setup complete.")

    if config["postbs"]:
        cloud_config.add_status_commands("cloud-init: Running postbs commands.")
        cloud_config.add_command("echo 'Starting custom postbs commands'")
        cloud_config.add_commands(config["postbs"])


def add_wait_for_cmd_ready_commands(cloud_config):
    cloud_config.add_status_commands("cloud-init: Waiting for CMDaemon to be ready.")
    cloud_config.add_command("chmod +x " + CMD_READY_PROBE_PATH)
    cloud_config.add_command(CMD_READY_PROBE_PATH)
    cloud_config.add_status_commands("cloud-init: CMDaemon is ready.")
